"""
accelerator.py

Base Hydra Structured Configs for defining various accelerator schemes. Uses a simple single inheritance structure.
"""
import os
from dataclasses import dataclass

from hydra.core.config_store import ConfigStore
from omegaconf import MISSING

# === Vanilla Accelerators (Deprecated; mostly for XLA code) ===


@dataclass
class AcceleratorConfig:
    accelerator: str = MISSING
    num_accelerators: int = MISSING
    num_workers: int = MISSING


@dataclass
class TPUv2OneConfig(AcceleratorConfig):
    accelerator = "tpu"
    num_accelerators = 1
    num_workers = 4


@dataclass
class TPUv2EightConfig(AcceleratorConfig):
    accelerator = "tpu"
    num_accelerators = 8
    num_workers = 4


@dataclass
class TPUv3OneConfig(AcceleratorConfig):
    accelerator = "tpu"
    num_accelerators = 1
    num_workers = 8


@dataclass
class TPUv3EightConfig(AcceleratorConfig):
    accelerator = "tpu"
    num_accelerators = 8
    num_workers = 8


# === GPU Default Config --> just set `num_workers`; `torchrun` takes care of the rest! ===
#   > Note :: Defaults to 1 GPU if WORLD_SIZE not set (e.g., not running with `torchrun`)


@dataclass
class TorchRunDefaultConfig(AcceleratorConfig):
    accelerator = "gpu"
    num_accelerators = int(os.environ["WORLD_SIZE"] if "WORLD_SIZE" in os.environ else 1)
    num_workers = 8


# Create a configuration group `accelerator` and populate with the above...
cs = ConfigStore.instance()
cs.store(group="accelerator", name="tpu-v2-1", node=TPUv2OneConfig)
cs.store(group="accelerator", name="tpu-v2-8", node=TPUv2EightConfig)
cs.store(group="accelerator", name="tpu-v3-1", node=TPUv3OneConfig)
cs.store(group="accelerator", name="tpu-v3-8", node=TPUv3EightConfig)

cs.store(group="accelerator", name="torchrun", node=TorchRunDefaultConfig)



"""
checkpointing.py

Core utility class for handling model/optimizer serialization & checkpointing -- including resume from checkpoint logic.

Support the following strategies:
    - (k, -1, -1) --> Keep only the most recent "k" epoch checkpoints
    - (k, m, -1) --> Keep the most recent "k" epoch checkpoints and *every* m epoch checkpoint
    - (k, m, s = 2500) --> Keep "k" and "m" subject to above, but also keep *s* step checkpoints for current epoch
"""
import logging
import os
import re
from collections import deque
from pathlib import Path
from typing import Any, Optional, Tuple

import torch
import torch.nn as nn
from torch.optim.optimizer import Optimizer

# Grab Logger
overwatch = logging.getLogger(__file__)


class FixedDeck(deque):
    def __init__(self, maxlen: int) -> None:
        super().__init__(maxlen=maxlen)

    def append(self, x: Any) -> Any:
        pop_value = None
        if self.__len__() == self.maxlen:
            pop_value = self.__getitem__(0)

        # Perform parent append and return popped value, if any!
        super().append(x)
        return pop_value


class CheckpointSaver:
    def __init__(self, strategy: Tuple[int, int, int], run_dir: str, is_rank_zero: bool = False) -> None:
        """
        Create a checkpoint saver with the provided strategy that saves to the given path.

        :param strategy: Strategy, following the (k, -1, -1) -- (k, m, -1) -- (k, m, s) description above.
        :param run_dir: Path to root of `run_dir`.
        :param is_rank_zero: Boolean whether this process is global zero (no-op if not)!
        """
        (self.k, self.m, self.s), self.run_dir, self.is_rank_zero = strategy, run_dir, is_rank_zero
        self.recents, self.intervals, self.step_checkpoints = FixedDeck(maxlen=self.k), set(), set()

        # If `self.s == -1` --> *Disable* step checkpoints (only at save end of epoch!)
        self.enable_step = self.s != -1

        # Create "checkpoints" subdirectory
        self.path = Path(run_dir) / "checkpoints"
        if self.is_rank_zero:
            os.makedirs(self.path, exist_ok=True)

            # Populate `step_checkpoints` on __init__ (if resuming *within* an epoch!)
            self.step_checkpoints.update([c for c in self.path.iterdir() if "local-epoch=" in str(c)])

        # Created Saver...
        overwatch.info(f"Created CheckpointSaver with `k = {self.k}` -- `m = {self.m}` -- s = {self.s}!")

    def save(
        self,
        epoch: int,
        is_local_step: bool,
        model: nn.Module,
        optimizer: Optimizer,
        duration: int,
        local_step: Optional[int] = None,
        train_loss: Optional[float] = None,
        val_loss: Optional[float] = None,
    ) -> None:
        """Performs a global zero save operation, unlinking stale checkpoints if necessary."""
        if not self.is_rank_zero:
            return

        # Check if saving a `local_step` (within an epoch) or if end of epoch...
        if self.enable_step and is_local_step and (local_step % self.s) == 0:
            step_checkpoint = self.path / f"local-epoch={epoch}-step={local_step}-t={duration}.pt"
            torch.save(
                {"model_state_dict": model.state_dict(), "optimizer_state_dict": optimizer.state_dict()}, step_checkpoint
            )

            # Update Relevant Trackers...
            self.step_checkpoints.add(step_checkpoint)

        elif not is_local_step:
            if train_loss is None and val_loss is None:
                checkpoint = self.path / f"epoch={epoch}-train=inf-val=inf-t={duration}.pt"
            else:
                checkpoint = self.path / f"epoch={epoch}-train={train_loss:.4f}-val={val_loss:.4f}-t={duration}.pt"
            torch.save(
                {"model_state_dict": model.state_dict(), "optimizer_state_dict": optimizer.state_dict()}, checkpoint
            )

            # Update Relevant Trackers
            if epoch % self.m == 0:
                self.intervals.add(checkpoint)

            # Remove all "step_checkpoints" now that we've made it to the end of an epoch!
            while len(self.step_checkpoints) > 0:
                os.remove(self.step_checkpoints.pop())

            # Add to recents & flush stale checkpoints...
            to_remove = self.recents.append(checkpoint)
            if to_remove is not None and to_remove not in self.intervals:
                os.remove(to_remove)


def do_resume(resume: bool, run_dir: str) -> Tuple[Optional[Path], int, int]:
    """Handle `resume` logic --> consists of retrieving checkpoint_path and epoch/step computation (if resuming)."""
    if not resume:
        # We're starting a fresh run --> return None for checkpoint_path, resume_epoch = 0, resume_step = 0
        return None, 0, 0

    # === Auto-Resume Logic ===
    # **IMPORTANT**: We're making a few assumptions on resuming that should eventually become explicit checks:
    #   - `accumulate_grad_batches` is exactly the same when resuming; this means:
    #       + `model_cfg.effective_bsz`, `model_cfg.fabric_bsz`, & `accelerator_cfg.num_accelerators` are the same!
    #   - The Weights & Biases directory `run_dir/wandb` only contains a *single run*
    #   - The `param_groups` in `optimizer.state_dict()` are exactly the same across resumes!
    #       + This means that (and generally should be true for resuming altogether) the architecture is the same!
    #   - The `cfg.seed` should be the same (again, should generally be true...)
    all_checkpoints_path, resume_checkpoint, resume_epoch, resume_step = Path(run_dir) / "checkpoints", None, 0, 0
    if all_checkpoints_path.exists() and any(all_checkpoints_path.iterdir()):
        # Parse out the latest "complete" epoch checkpoint, as well as any "local step" checkpoints...
        checkpoints = list(all_checkpoints_path.iterdir())
        complete_checkpoint, complete_epoch = max(
            [
                (c, int(re.search("epoch=(.+?)-train", c.name).group(1)))
                for c in checkpoints
                if "local-epoch=" not in str(c)
            ],
            key=lambda x: x[1],
        )

        # Case 1 :: We have "local step" checkpoints --> will always override any "full epoch" checkpoints...
        local = [
            (
                c,
                int(re.search("local-epoch=(.+?)-step", c.name).group(1)),
                int(re.search("step=(.+?)[.-]", c.name).group(1)),
            )
            for c in checkpoints
            if "local-epoch=" in str(c)
        ]
        if len(local) > 0:
            # Parse out (epoch, "highest" step) + assert no great "full epoch" checkpoint exists!
            resume_checkpoint, resume_epoch, resume_step = max(local, key=lambda x: x[1:])
            assert resume_epoch == complete_epoch, "Epoch mismatch in `resume` from local_step!"

        # Case 2 :: Otherwise, we're just going to start with the last "complete" epoch
        else:
            resume_checkpoint, resume_epoch = complete_checkpoint, complete_epoch

    return resume_checkpoint, resume_epoch, resume_step



"""
utilities.py

General utilities for randomness, distributed training, and miscellaneous checks in PyTorch.

=== Randomness ===

Random `seed_everything` functionality is taken directly from PyTorch-Lighting:
    > Ref: https://github.com/PyTorchLightning/pytorch-lightning/blob/master/pytorch_lightning/utilities/seed.py

This is pretty important to get right if we're every randomly generating our masks (or prefix dropout) inside our
Dataset __getitem__() with multiple workers... if not handled properly, we will get repeated augmentations anytime
we inject randomness from non-PyTorch sources (e.g., numpy, random)!
    > Ref: https://tanelp.github.io/posts/a-bug-that-plagues-thousands-of-open-source-ml-projects/

=== Distributed / DDP Training ====

Utilities provide a standard API across single-GPU/multi-GPU/multi-node training. Assumes that code is running with
one of the following strategies:
    - Single Process (on CPU?, GPU)
    - DDP (GPU, Multi-Node GPU) --> uses the `torchrun`/`torch.distributed` API & semantics

Key Terminology
    -> World Size :: Total number of processes distributed over (# nodes x # devices) -- assumed homogenous!
    -> Rank :: Integer index of current process in the total world size
    -> Local Rank :: Local index on given node in [0, Devices per Node]
"""
import os
import random
from typing import Callable, Iterator, Optional, TypeVar

import numpy as np
import torch
from torch.utils.data import Dataset
from torch.utils.data.distributed import DistributedSampler

T_co = TypeVar("T_co", covariant=True)


# === Randomness ===


def worker_init_function(worker_id: int) -> None:
    """
    Borrowed directly from PyTorch-Lightning; inspired by this issue comment in the PyTorch repo:
        > Ref: https://github.com/pytorch/pytorch/issues/5059#issuecomment-817392562

    Intuition: You can think of the seed sequence spawn function as a "janky" torch.Generator() or jax.PRNGKey that
    you can run iterative splitting on to get new (predictable) randomness.

    :param worker_id: Identifier for the given worker [0, num_workers) for the Dataloader in question.
    """
    # Get current `rank` (if running distributed) and `process_seed`
    global_rank, process_seed = int(os.environ["LOCAL_RANK"]), torch.initial_seed()

    # Back out the "base" (original) seed - the per-worker seed is set in PyTorch:
    #   > https://pytorch.org/docs/stable/data.html#data-loading-randomness
    base_seed = process_seed - worker_id

    # "Magic" code --> basically creates a seed sequence that mixes different "sources" and seeds every library...
    seed_seq = np.random.SeedSequence([base_seed, worker_id, global_rank])

    # Use 128 bits (4 x 32-bit words) to represent seed --> generate_state(k) produces a `k` element array!
    np.random.seed(seed_seq.generate_state(4))

    # Spawn distinct child sequences for PyTorch (reseed) and stdlib random
    torch_seed_seq, random_seed_seq = seed_seq.spawn(2)

    # Torch Manual seed takes 64 bits (so just specify a dtype of uint64
    torch.manual_seed(torch_seed_seq.generate_state(1, dtype=np.uint64)[0])

    # Use 128 Bits for `random`, but express as integer instead of as an array
    random_seed = (random_seed_seq.generate_state(2, dtype=np.uint64).astype(list) * [1 << 64, 1]).sum()
    random.seed(random_seed)


def set_global_seed(seed: int, get_worker_init_fn: bool = False) -> Optional[Callable[[int], None]]:
    """Sets seed for all randomness libraries (mostly random, numpy, torch) and produces a `worker_init_fn`"""
    assert np.iinfo(np.uint32).min < seed < np.iinfo(np.uint32).max, "Seed outside the np.uint32 bounds!"

    # Set Seed as an Environment Variable
    os.environ["EXPERIMENT_GLOBAL_SEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

    return worker_init_function if get_worker_init_fn else None


# === Distributed Training ===


class ResumeableDistributedSampler(DistributedSampler):
    def __init__(
        self,
        seen_examples: int,
        resume_epoch: int,
        dataset: Dataset,
        num_replicas: int,
        rank: int,
        shuffle: bool = True,
        seed: int = 0,
    ) -> None:
        super().__init__(dataset, num_replicas=num_replicas, rank=rank, shuffle=shuffle, seed=seed)
        self.seen_examples, self.resume_epoch, self.do_resume = seen_examples, resume_epoch, True

        # Set `seen_examples_per_replica` --> this is necessary for when we re-wrap the iterator in self.__iter__()
        #   > Note: `seen_examples` is across _all_ replicas --> so divide!
        self.seen_examples_per_replica = self.seen_examples // self.num_replicas

    def __iter__(self) -> Iterator[T_co]:
        epoch_iterator = super().__iter__()
        if self.do_resume:
            # Unpack iterator --> list, slice off the first `seen_examples_per_replica` examples, and re-wrap!
            leftover_idxs = list(epoch_iterator)[self.seen_examples_per_replica :]
            return iter(leftover_idxs)
        else:
            return epoch_iterator

    def __len__(self) -> int:
        if self.do_resume:
            # Remove the "seen" sample from self.num_samples; num_samples is *per replica*!
            return self.num_samples - self.seen_examples_per_replica
        else:
            return self.num_samples

    def set_epoch(self, epoch: int) -> None:
        # If epoch != self.resume_epoch --> we're in "regular DistributedSampler" mode (just a wrapper class)
        #   > Intuition: We should *only* truncate examples on the first epoch upon resuming!
        self.epoch = epoch
        if self.epoch != self.resume_epoch:
            self.do_resume = False



"""
instantiate.py

Simple wrapping script for instantiating a core Voltron/reproduction model and configuring the torch.Optimizer for DDP
pretraining. Meant to be modular and extensible!
"""
from typing import Callable, Tuple

import torch.nn as nn
from torch.optim import Optimizer

from voltron.conf import DatasetConfig, ModelConfig

from .core.vcond import VCond
from .core.vdual import VDual
from .core.vgen import VGen
from .reproductions.vmvp import VMVP
from .reproductions.vr3m import VR3M
from .reproductions.vrn3m import VRN3M


def get_model_optimizer(
    model_cfg: ModelConfig, dataset_cfg: DatasetConfig
) -> Tuple[nn.Module, Optimizer, Callable[[int, float], float]]:
    """Switch on `model_cfg.arch` --> instantiate the correct nn.Module and Optimizer (on CPU/default device)."""

    # Data-Locked Reproductions
    if model_cfg.arch == "v-mvp":
        model = VMVP(
            resolution=dataset_cfg.resolution,
            patch_size=model_cfg.patch_size,
            encoder_depth=model_cfg.encoder_depth,
            encoder_embed_dim=model_cfg.encoder_embed_dim,
            encoder_n_heads=model_cfg.encoder_n_heads,
            decoder_depth=model_cfg.decoder_depth,
            decoder_embed_dim=model_cfg.decoder_embed_dim,
            decoder_n_heads=model_cfg.decoder_n_heads,
            optimizer=model_cfg.optimizer,
            schedule=model_cfg.schedule,
            base_lr=model_cfg.base_lr,
            min_lr=model_cfg.min_lr,
            effective_bsz=model_cfg.effective_bsz,
            betas=model_cfg.betas,
            weight_decay=model_cfg.weight_decay,
            warmup_epochs=dataset_cfg.warmup_epochs,
            max_epochs=dataset_cfg.max_epochs,
            mlp_ratio=model_cfg.mlp_ratio,
            norm_pixel_loss=model_cfg.norm_pixel_loss,
        )

    elif model_cfg.arch == "v-r3m":
        model = VR3M(
            resolution=dataset_cfg.resolution,
            patch_size=model_cfg.patch_size,
            depth=model_cfg.depth,
            embed_dim=model_cfg.embed_dim,
            n_heads=model_cfg.n_heads,
            language_model=model_cfg.language_model,
            hf_cache=model_cfg.hf_cache,
            language_dim=model_cfg.language_dim,
            reward_dim=model_cfg.reward_dim,
            n_negatives=model_cfg.n_negatives,
            lang_reward_weight=model_cfg.lang_reward_weight,
            tcn_weight=model_cfg.tcn_weight,
            l1_weight=model_cfg.l1_weight,
            l2_weight=model_cfg.l2_weight,
            optimizer=model_cfg.optimizer,
            schedule=model_cfg.schedule,
            lr=model_cfg.lr,
            min_lr=model_cfg.min_lr,
            warmup_epochs=dataset_cfg.warmup_epochs,
            max_epochs=dataset_cfg.max_epochs,
            mlp_ratio=model_cfg.mlp_ratio,
        )

    elif model_cfg.arch == "v-rn3m":
        model = VRN3M(
            resolution=dataset_cfg.resolution,
            fc_dim=model_cfg.fc_dim,
            language_model=model_cfg.language_model,
            hf_cache=model_cfg.hf_cache,
            language_dim=model_cfg.language_dim,
            reward_dim=model_cfg.reward_dim,
            n_negatives=model_cfg.n_negatives,
            lang_reward_weight=model_cfg.lang_reward_weight,
            tcn_weight=model_cfg.tcn_weight,
            l1_weight=model_cfg.l1_weight,
            l2_weight=model_cfg.l2_weight,
            optimizer=model_cfg.optimizer,
            lr=model_cfg.lr,
        )

    # Voltron Models
    elif model_cfg.arch == "v-cond":
        model = VCond(
            resolution=dataset_cfg.resolution,
            patch_size=model_cfg.patch_size,
            encoder_depth=model_cfg.encoder_depth,
            encoder_embed_dim=model_cfg.encoder_embed_dim,
            encoder_n_heads=model_cfg.encoder_n_heads,
            decoder_depth=model_cfg.decoder_depth,
            decoder_embed_dim=model_cfg.decoder_embed_dim,
            decoder_n_heads=model_cfg.decoder_n_heads,
            language_model=model_cfg.language_model,
            hf_cache=model_cfg.hf_cache,
            language_dim=model_cfg.language_dim,
            optimizer=model_cfg.optimizer,
            schedule=model_cfg.schedule,
            base_lr=model_cfg.base_lr,
            min_lr=model_cfg.min_lr,
            effective_bsz=model_cfg.effective_bsz,
            betas=model_cfg.betas,
            weight_decay=model_cfg.weight_decay,
            warmup_epochs=dataset_cfg.warmup_epochs,
            max_epochs=dataset_cfg.max_epochs,
            mlp_ratio=model_cfg.mlp_ratio,
            norm_pixel_loss=model_cfg.norm_pixel_loss,
        )

    elif model_cfg.arch == "v-dual":
        model = VDual(
            resolution=dataset_cfg.resolution,
            patch_size=model_cfg.patch_size,
            encoder_depth=model_cfg.encoder_depth,
            encoder_embed_dim=model_cfg.encoder_embed_dim,
            encoder_n_heads=model_cfg.encoder_n_heads,
            decoder_depth=model_cfg.decoder_depth,
            decoder_embed_dim=model_cfg.decoder_embed_dim,
            decoder_n_heads=model_cfg.decoder_n_heads,
            language_model=model_cfg.language_model,
            hf_cache=model_cfg.hf_cache,
            language_dim=model_cfg.language_dim,
            optimizer=model_cfg.optimizer,
            schedule=model_cfg.schedule,
            base_lr=model_cfg.base_lr,
            min_lr=model_cfg.min_lr,
            effective_bsz=model_cfg.effective_bsz,
            betas=model_cfg.betas,
            weight_decay=model_cfg.weight_decay,
            warmup_epochs=dataset_cfg.warmup_epochs,
            max_epochs=dataset_cfg.max_epochs,
            mlp_ratio=model_cfg.mlp_ratio,
            norm_pixel_loss=model_cfg.norm_pixel_loss,
        )

    elif model_cfg.arch == "v-gen":
        model = VGen(
            resolution=dataset_cfg.resolution,
            patch_size=model_cfg.patch_size,
            encoder_depth=model_cfg.encoder_depth,
            encoder_embed_dim=model_cfg.encoder_embed_dim,
            encoder_n_heads=model_cfg.encoder_n_heads,
            decoder_depth=model_cfg.decoder_depth,
            decoder_embed_dim=model_cfg.decoder_embed_dim,
            decoder_n_heads=model_cfg.decoder_n_heads,
            language_model=model_cfg.language_model,
            hf_cache=model_cfg.hf_cache,
            language_dim=model_cfg.language_dim,
            max_lang_len=dataset_cfg.max_lang_len,
            vocab_size=model_cfg.vocab_size,
            mae_weight=model_cfg.mae_weight,
            lm_weight=model_cfg.lm_weight,
            optimizer=model_cfg.optimizer,
            schedule=model_cfg.schedule,
            base_lr=model_cfg.base_lr,
            min_lr=model_cfg.min_lr,
            effective_bsz=model_cfg.effective_bsz,
            betas=model_cfg.betas,
            weight_decay=model_cfg.weight_decay,
            warmup_epochs=dataset_cfg.warmup_epochs,
            max_epochs=dataset_cfg.max_epochs,
            mlp_ratio=model_cfg.mlp_ratio,
            norm_pixel_loss=model_cfg.norm_pixel_loss,
        )

    else:
        raise ValueError(f"Model Architecture `{model_cfg.arch}` is not implemented!")

    # Configure Optimizer --> on same device (CPU)
    optimizer, update_lr = model.configure_optimizer()

    return model, optimizer, update_lr



"""
xpretrain.py

(The `x` prefix indicates this is a script geared for XLA/TPU backends *only*)!

Reference script for PyTorch XLA (TPU-based) pretraining on the Something-Something-v2 dataset; this is
mostly for completeness =>> the hope is that the regular `pretrain.py` script is more general and maintained.

Focuses on multi-TPU (XLA) training --> but also supports single-core TPU training, as the default distributed mp.spawn
behavior just collapses into a single thread! Loads and preprocesses dataset, instantiates a model, and runs training.

Run with `python examples/xla-reference/xpretrain.py` (will use the configuration specified by `DEFAULTS` below).
"""
import os
import re
import time
from collections import deque
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

import hydra
import jsonlines
import numpy as np
import torch
import torch_xla.core.xla_model as xm
import torch_xla.debug.metrics as met
import torch_xla.distributed.parallel_loader as parallel
import wandb
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING, OmegaConf
from torch.utils.data import DataLoader
from tqdm import tqdm

from voltron.conf import AcceleratorConfig, DatasetConfig, ModelConfig, TrackingConfig
from voltron.datasets.v1.stream_datasets import get_epoch_datasets
from voltron.models import VMVP, VR3M, VRN3M, VCond, VDual, VGen
from voltron.overwatch import OverwatchRich
from voltron.util.v1.checkpointing import XLACheckpointSaver
from voltron.util.v1.distributed import ResumeableDistributedSampler
from voltron.util.v1.random import set_global_seed
from voltron.util.v1.xla_logger import (
    log_epoch_end_update,
    log_vcond_train_update,
    log_vdual_train_update,
    log_vgen_train_update,
    log_vmvp_train_update,
    log_vr3m_train_update,
    log_vrn3m_train_update,
)

# Set Defaults (Hydra w/ Structured Configs)
DEFAULTS = [
    "_self_",
    {"model": "v-cond"},
    {"dataset": "sth-sth-v2"},
    {"accelerator": "tpu-v3-8"},
    {"tracking": "voltron-tracking"},
    {"override hydra/job_logging": "overwatch_rich"},
]


@dataclass
class PretrainConfig:
    # fmt: off
    defaults: List[Any] = field(default_factory=lambda: DEFAULTS)
    hydra: Dict[str, Any] = field(default_factory=lambda: {
        "run": {"dir": "./runs/train/${model.identifier}+dataset-${dataset.name}"}
    })

    # Command Line Arguments
    run_id: Optional[str] = None                                        # Run ID for Logging
    seed: int = 21                                                      # Random Seed (for reproducibility)

    # Resume / Debug Behavior
    resume: bool = True                                                 # Whether to resume an existing run...
    resume_epoch: Optional[int] = None                                  # Epoch to resume (if auto-resuming)...
    checkpoint_path: Optional[str] = None                               # Path to the specific checkpoint to load!
    wandb_resume_id: Optional[str] = None                               # W&B Run ID for `resume` behavior...

    # Composable / Structured Arguments
    model: ModelConfig = MISSING                                        # Model architecture for pretraining
    dataset: DatasetConfig = MISSING                                    # List of datasets for pretraining
    accelerator: AcceleratorConfig = MISSING                            # Accelerator configuration
    tracking: TrackingConfig = MISSING                                  # Run/experiment tracking configuration
    # fmt: on


# Hydra Setup :: Retrieve ConfigStore (Singleton) & Register Components
cs = ConfigStore.instance()
cs.store(group="hydra/job_logging", name="overwatch_rich", node=OverwatchRich)  # Annoying - configure logger for Hydra
cs.store(name="config", node=PretrainConfig)


# ruff: noqa: C901
def xpretrain(cfg: PretrainConfig) -> None:
    # Identify if `is_rank_zero` --> We only log from the rank zero process!
    is_rank_zero = xm.is_master_ordinal(local=False)
    xm.master_print("Voltron Training :: Assembling the Legendary Defender...")

    # Create Unique Run Name -- if `resume = True` we assume same "run_id"
    run_id = cfg.run_id
    if run_id is None:
        run_id = run_dir = f"{cfg.model.identifier}+{cfg.dataset.name}-x{cfg.seed}"
        cfg.run_id = run_id
    else:
        cfg.run_id = run_dir = run_id

    if is_rank_zero:
        os.makedirs(run_dir, exist_ok=True)

    xm.master_print(
        '\t=>> "If you get too worried about what could go wrong, you might miss a chance to do something great."'
    )

    # Set Randomness, get DataLoader worker initialization function (to ensure any random augmentations!)
    worker_init_fn = set_global_seed(cfg.seed)

    # Model Initialization Logic
    xm.master_print("Initializing Model and Placing on Different Devices...")
    if cfg.model.arch == "v-mvp":
        xm.master_print(f"Initializing MVP variant `{cfg.model.identifier}`")
        model = VMVP(
            resolution=cfg.dataset.resolution,
            patch_size=cfg.model.patch_size,
            encoder_depth=cfg.model.encoder_depth,
            encoder_embed_dim=cfg.model.encoder_embed_dim,
            encoder_n_heads=cfg.model.encoder_n_heads,
            decoder_depth=cfg.model.decoder_depth,
            decoder_embed_dim=cfg.model.decoder_embed_dim,
            decoder_n_heads=cfg.model.decoder_n_heads,
            optimizer=cfg.model.optimizer,
            schedule=cfg.model.schedule,
            base_lr=cfg.model.base_lr,
            min_lr=cfg.model.min_lr,
            effective_bsz=cfg.model.effective_bsz,
            betas=cfg.model.betas,
            weight_decay=cfg.model.weight_decay,
            warmup_epochs=cfg.dataset.warmup_epochs,
            max_epochs=cfg.dataset.max_epochs,
            mlp_ratio=cfg.model.mlp_ratio,
            norm_pixel_loss=cfg.model.norm_pixel_loss,
        )

    elif cfg.model.arch == "v-r3m":
        xm.master_print(f"Initializing R3M (ViT) Variant `{cfg.model.identifier}`")
        model = VR3M(
            resolution=cfg.dataset.resolution,
            patch_size=cfg.model.patch_size,
            depth=cfg.model.depth,
            embed_dim=cfg.model.embed_dim,
            n_heads=cfg.model.n_heads,
            language_model=cfg.model.language_model,
            hf_cache=cfg.model.hf_cache,
            language_dim=cfg.model.language_dim,
            reward_dim=cfg.model.reward_dim,
            n_negatives=cfg.model.n_negatives,
            lang_reward_weight=cfg.model.lang_reward_weight,
            tcn_weight=cfg.model.tcn_weight,
            l1_weight=cfg.model.l1_weight,
            l2_weight=cfg.model.l2_weight,
            optimizer=cfg.model.optimizer,
            schedule=cfg.model.schedule,
            lr=cfg.model.lr,
            min_lr=cfg.model.min_lr,
            warmup_epochs=cfg.dataset.warmup_epochs,
            max_epochs=cfg.dataset.max_epochs,
            mlp_ratio=cfg.model.mlp_ratio,
        )

    elif cfg.model.arch == "v-rn3m":
        xm.master_print(f"Intializing R3M (ResNet) Variant `{cfg.model.identifier}`")
        model = VRN3M(
            resolution=cfg.dataset.resolution,
            fc_dim=cfg.model.fc_dim,
            language_model=cfg.model.language_model,
            hf_cache=cfg.model.hf_cache,
            language_dim=cfg.model.language_dim,
            reward_dim=cfg.model.reward_dim,
            n_negatives=cfg.model.n_negatives,
            lang_reward_weight=cfg.model.lang_reward_weight,
            tcn_weight=cfg.model.tcn_weight,
            l1_weight=cfg.model.l1_weight,
            l2_weight=cfg.model.l2_weight,
            optimizer=cfg.model.optimizer,
            lr=cfg.model.lr,
        )

    elif cfg.model.arch == "v-cond":
        xm.master_print(f"Initializing Voltron V-Cond variant `{cfg.model.identifier}`")
        model = VCond(
            resolution=cfg.dataset.resolution,
            patch_size=cfg.model.patch_size,
            encoder_depth=cfg.model.encoder_depth,
            encoder_embed_dim=cfg.model.encoder_embed_dim,
            encoder_n_heads=cfg.model.encoder_n_heads,
            decoder_depth=cfg.model.decoder_depth,
            decoder_embed_dim=cfg.model.decoder_embed_dim,
            decoder_n_heads=cfg.model.decoder_n_heads,
            language_model=cfg.model.language_model,
            hf_cache=cfg.model.hf_cache,
            language_dim=cfg.model.language_dim,
            optimizer=cfg.model.optimizer,
            schedule=cfg.model.schedule,
            base_lr=cfg.model.base_lr,
            min_lr=cfg.model.min_lr,
            effective_bsz=cfg.model.effective_bsz,
            betas=cfg.model.betas,
            weight_decay=cfg.model.weight_decay,
            warmup_epochs=cfg.dataset.warmup_epochs,
            max_epochs=cfg.dataset.max_epochs,
            mlp_ratio=cfg.model.mlp_ratio,
            norm_pixel_loss=cfg.model.norm_pixel_loss,
        )

    elif cfg.model.arch == "v-dual":
        xm.master_print(f"Initializing Voltron V-Dual variant `{cfg.model.identifier}`")
        model = VDual(
            resolution=cfg.dataset.resolution,
            patch_size=cfg.model.patch_size,
            encoder_depth=cfg.model.encoder_depth,
            encoder_embed_dim=cfg.model.encoder_embed_dim,
            encoder_n_heads=cfg.model.encoder_n_heads,
            decoder_depth=cfg.model.decoder_depth,
            decoder_embed_dim=cfg.model.decoder_embed_dim,
            decoder_n_heads=cfg.model.decoder_n_heads,
            language_model=cfg.model.language_model,
            hf_cache=cfg.model.hf_cache,
            language_dim=cfg.model.language_dim,
            optimizer=cfg.model.optimizer,
            schedule=cfg.model.schedule,
            base_lr=cfg.model.base_lr,
            min_lr=cfg.model.min_lr,
            effective_bsz=cfg.model.effective_bsz,
            betas=cfg.model.betas,
            weight_decay=cfg.model.weight_decay,
            warmup_epochs=cfg.dataset.warmup_epochs,
            max_epochs=cfg.dataset.max_epochs,
            mlp_ratio=cfg.model.mlp_ratio,
            norm_pixel_loss=cfg.model.norm_pixel_loss,
        )

    elif cfg.model.arch == "v-gen":
        xm.master_print(f"Initializing Voltron V-Gen variant `{cfg.model.identifier}`")
        model = VGen(
            resolution=cfg.dataset.resolution,
            patch_size=cfg.model.patch_size,
            encoder_depth=cfg.model.encoder_depth,
            encoder_embed_dim=cfg.model.encoder_embed_dim,
            encoder_n_heads=cfg.model.encoder_n_heads,
            decoder_depth=cfg.model.decoder_depth,
            decoder_embed_dim=cfg.model.decoder_embed_dim,
            decoder_n_heads=cfg.model.decoder_n_heads,
            language_model=cfg.model.language_model,
            hf_cache=cfg.model.hf_cache,
            language_dim=cfg.model.language_dim,
            max_lang_len=cfg.dataset.max_lang_len,
            vocab_size=cfg.model.vocab_size,
            mae_weight=cfg.model.mae_weight,
            lm_weight=cfg.model.lm_weight,
            optimizer=cfg.model.optimizer,
            schedule=cfg.model.schedule,
            base_lr=cfg.model.base_lr,
            min_lr=cfg.model.min_lr,
            effective_bsz=cfg.model.effective_bsz,
            betas=cfg.model.betas,
            weight_decay=cfg.model.weight_decay,
            warmup_epochs=cfg.dataset.warmup_epochs,
            max_epochs=cfg.dataset.max_epochs,
            mlp_ratio=cfg.model.mlp_ratio,
            norm_pixel_loss=cfg.model.norm_pixel_loss,
        )

    else:
        raise NotImplementedError(f"Model Architecture `{cfg.model.arch}` is not supported!")

    # We use gradient accumulation to honor the effective batch size specified...
    assert cfg.model.effective_bsz % cfg.model.device_bsz == 0, "Device bsz must evenly divide effective bsz!"
    accumulate_grad_batches = cfg.model.effective_bsz // cfg.model.device_bsz // xm.xrt_world_size()
    xm.master_print(
        f"Running `{cfg.model.identifier}` model w/ Effective Batch Size of `{cfg.model.effective_bsz}`, "
        f"Per-Device Batch Size of `{cfg.model.device_bsz}`, "
        f"Distributed World Size of `{xm.xrt_world_size()}` and `{accumulate_grad_batches}` Accumulation Steps"
    )

    # If Resuming =>> Load Model from Checkpoint
    start_checkpoint, start_epoch, start_step = None, 0, 0
    if cfg.resume:
        # **IMPORTANT**: We're making a few assumptions on resuming that should eventually become explicit checks:
        #   - `accumulate_grad_batches` is exactly the same when resuming; this means:
        #       + `cfg.model.effective_bsz`, `cfg.model.device_bsz`, & `cfg.accelerator.num_accelerators` are the same!
        #   - The Weights & Biases directory `run_dir/wandb` only contains a *single run*
        #   - The `param_groups` in `optimizer.state_dict()` are exactly the same across resumes!
        #       + This means that (and generally should be true for resuming altogether) the architecture is the same!
        #   - The `cfg.seed` should be the same (again, should generally be true...)
        if cfg.checkpoint_path is None:
            xm.master_print("Resuming :: Attempting to Automatically Load Checkpoint -- Searching!")
            checkpoint_path = Path(run_dir) / "checkpoints"
            if checkpoint_path.exists() and any(checkpoint_path.iterdir()):
                # Parse out the latest "complete" epoch checkpoint, as well as any "local step" checkpoints...
                checkpoints = list(checkpoint_path.iterdir())
                complete_checkpoint, complete_epoch = max(
                    [
                        (c, int(re.search("epoch=(.+?)-train", c.name).group(1)))
                        for c in checkpoints
                        if "local-epoch=" not in str(c)
                    ],
                    key=lambda x: x[1],
                )

                # Case 1 :: We have "local step" checkpoints --> will always override any "full epoch" checkpoints...
                local = [
                    (
                        c,
                        int(re.search("local-epoch=(.+?)-step", c.name).group(1)),
                        int(re.search("step=(.+?)[.-]", c.name).group(1)),
                    )
                    for c in checkpoints
                    if "local-epoch=" in str(c)
                ]
                if len(local) > 0:
                    # Parse out (epoch, "highest" step) + assert no great "full epoch" checkpoint exists!
                    start_checkpoint, start_epoch, start_step = max(local, key=lambda x: x[1:])
                    assert start_epoch == complete_epoch, "Epoch mismatch in `resume` from local_step!"

                # Case 2 :: Otherwise, we're just going to start with the last "complete" epoch...
                else:
                    start_checkpoint, start_epoch = complete_checkpoint, complete_epoch

            else:
                xm.master_print("No Checkpoints Found -- Starting Run from Scratch!")

        else:
            xm.master_print(f"Resuming :: Loading from Checkpoint `{cfg.checkpoint_path}`...")
            start_checkpoint = cfg.checkpoint_path

        # Actually Load the Checkpoint State!
        if start_checkpoint is not None:
            xm.master_print(f"Resuming :: Loading Model & Optimizer State Dictionaries from `{start_checkpoint}`")
            checkpoint = torch.load(str(start_checkpoint))
            model_state_dict, optimizer_state_dict = checkpoint
            model.load_state_dict(model_state_dict)

    # Logging / W&B Handling
    if is_rank_zero:
        xm.master_print("Initializing Weights & Biases + JSONL + Checkpoint Saver on Rank Zero ONLY...")
        tags = None
        if cfg.tracking.tags is None:
            tags = [cfg.model.identifier, cfg.dataset.name, "pretraining"]

        # W&B Initialize & Log all Hyperparameters (Only on ordinal 0)
        wandb_resume_id = None
        if cfg.resume and cfg.wandb_resume_id is None:
            xm.master_print("Resuming :: Attempting to Automatically Load W&B Resume ID -- Searching!")
            wandb_path = Path("wandb")
            if wandb_path.exists() and any((wandb_path / "latest-run").iterdir()):
                # Parse out the unique resume_id from the `.wandb` file...
                wandb_fns = [f.name for f in (wandb_path / "latest-run").iterdir() if str(f).endswith(".wandb")]
                assert len(wandb_fns) == 1, f"There should only be 1 `.wandb` file... found {len(wandb_fns)}!"

                # Regex match on `run-{id}.wandb`...
                wandb_resume_id = re.search("run-(.+?).wandb", wandb_fns[0]).group(1)

            # Otherwise, assert that we're starting from scratch!
            else:
                assert start_checkpoint is None, "Trying to restart a run from checkpoint without a valid W&B ID!"

        elif cfg.resume:
            xm.master_print(f"Resuming :: Using Specified W&B Resume ID = `{cfg.wandb_resume_id}`")
            wandb_resume_id = cfg.wandb_resume_id

        # Initialize Weights & Biases
        xm.master_print(f"W&B Resume is {cfg.resume} w/ W&B Resume ID = {wandb_resume_id}!")
        wandb.init(
            project=cfg.tracking.project,
            entity=cfg.tracking.entity,
            config=cfg,
            name=run_id,
            dir=f"{os.getcwd()}" if cfg.tracking.directory is None else cfg.tracking.directory,
            tags=tags,
            notes=cfg.tracking.notes,
            resume="allow" if start_checkpoint is not None else False,
            id=wandb_resume_id,
            # Weird line because PT-TPU VMs don't come with a working install of Tensorflow...
            settings=wandb.Settings(_disable_stats=True),
        )

        # Initialize JSONL Logger (append only mode) --> last "global step" will always take precedence.
        with jsonlines.open(f"{run_id}.jsonl", mode="a", sort_keys=True) as js_logger:
            js_logger.write(
                {
                    "run_id": run_id,
                    "start_time": datetime.now().strftime("%m-%d-%H:%M"),
                    "hparams": OmegaConf.to_container(cfg),
                }
            )

    # Rank Zero Node will take time to spin up the loggers & checkpointer... might as well rendezvous?
    xm.rendezvous("Logging...")

    # === Here Be Dragons ===
    # Time to handle device placement -- Note - example code doesn't specify device idx - why not?
    #   > https://github.com/pytorch/xla/blob/3c0d68da07702995a592ea70f27868cd76fa0755/test/test_train_mp_mnist.py#L114
    #   > Results in printing [xla:0] and [xla:1] a bunch... no [xla:2-7]? This feels bad...?
    #
    #   |=> Debugging Try: `xm.xla_device(n=xm.get_ordinal()) ---> hangs completely?
    #   +=> *ANSWER*: https://github.com/pytorch/xla/issues/2345#issuecomment-657114819
    #       >> "Make no assumptions and don't try to build them manually..."
    device = xm.xla_device()
    model = model.train().to(device)
    optimizer, update_lr = model.configure_optimizer()
    global_step, train_losses, lrs, start_time, resume_time = 0, deque(maxlen=128), [], time.time(), 0

    # If resuming (valid `start_checkpoint`) -- patch the optimizer state dictionary, and load!
    if start_checkpoint is not None:
        patched_optimizer_state_dict = {
            "state": optimizer_state_dict,
            "param_groups": optimizer.state_dict()["param_groups"],
        }
        optimizer.load_state_dict(patched_optimizer_state_dict)

    # Create step timing...
    step_times, step_start_time = deque(maxlen=128), time.time()

    # Create Model/Architecture-Specific Trackers...
    if cfg.model.arch == "v-mvp":
        reconstruction_losses = deque(maxlen=128)

    elif cfg.model.arch in {"v-r3m", "v-rn3m"}:
        tcn_losses, reward_losses, l1_losses, l2_losses = [deque(maxlen=128) for _ in range(4)]
        tcn_accuracies, reward_accuracies = [deque(maxlen=128) for _ in range(2)]

    elif cfg.model.arch == "v-cond":
        reconstruction_losses = deque(maxlen=128)

    elif cfg.model.arch == "v-dual":
        reconstruction_losses = deque(maxlen=128)
        zero_reconstruction, k_reconstruction = deque(maxlen=128), deque(maxlen=128)

    elif cfg.model.arch == "v-gen":
        reconstruction_losses, lm_losses, lm_ppl = deque(maxlen=128), deque(maxlen=128), deque(maxlen=128)
        zero_reconstruction, k_reconstruction = deque(maxlen=128), deque(maxlen=128)

    else:
        raise NotImplementedError(f"Trackers for Model `{cfg.model.arch}` not implemented!")

    # 0th Checkpoint - Pull out optimizer state explicitly (`groups` are not serializable & can easily be replicated)
    saver = XLACheckpointSaver(cfg.tracking.checkpoint_strategy, run_dir, cfg.accelerator.accelerator)
    if start_checkpoint is None and start_epoch == 0:
        xm.master_print("Saving 0th Epoch Checkpoint...")
        saver.save(
            epoch=0, is_local_step=False, model=model, optimizer=optimizer, duration=0, train_loss=None, val_loss=None
        )

    # Run on all processes --> retrieve "0th epoch" dataset!
    #   =>> Important, ensures data is locked across models, for the given epoch!
    xm.master_print(f"Retrieving Dataset `{cfg.dataset.name}` prepared for `{cfg.model.arch}`!")
    train_dataset, val_dataset = get_epoch_datasets(
        0,
        cfg.dataset.name,
        cfg.dataset.normalization,
        cfg.model.arch,
        cfg.dataset.stream,
        cfg.dataset.artifact_path,
        cfg.dataset.stream_prefix,
        cfg.model.data_modality,
        cfg.model.get("lang_dropout", None),
        cfg.model.get("gen_ratio", None),
    )

    # Loading Datasets might take time... rendezvous to be safe
    xm.rendezvous("Retrieved Datasets...")

    # Iterate through Epochs, Evaluating at the end of each Training Epoch!
    #   >> Everything in this loop should happen across all workers, except for the logging (ordinal 0)!
    xm.master_print("Starting Training Loop...")
    for epoch in range(start_epoch, cfg.dataset.max_epochs):
        xm.master_print(f"\t[Epoch {epoch}] Building Distributed Sampler & DataLoaders...")
        train_dataset.set_epoch(epoch)

        # ResumeableDistributedSampler operates at over *examples* --> start_step (full_batches) * effective_bsz
        seen_examples = start_step * cfg.model.effective_bsz
        train_sampler = ResumeableDistributedSampler(
            seen_examples,
            start_epoch,
            train_dataset,
            num_replicas=xm.xrt_world_size(),
            rank=xm.get_ordinal(),
            shuffle=True,
            seed=cfg.seed,
        )

        # Set epoch appropriately for the `train_sampler` --> necessary to trigger "normal" logic!
        train_sampler.set_epoch(epoch)

        train_dataloader = DataLoader(
            train_dataset,
            batch_size=cfg.model.device_bsz,
            sampler=train_sampler,
            shuffle=False if train_sampler else True,
            num_workers=cfg.accelerator.num_workers,
            drop_last=True,
            worker_init_fn=worker_init_fn,
            prefetch_factor=4,
        )

        # NOTE :: We're not sharding the Validation set --> *everybody* will run forward passes on the *same* data!
        #   > We will have to reduce_mesh() later... unclear why, but the torch_xla folks seem keen on it; might lead to
        #   > weird rendezvous/hang issues if Validation is big enough...
        val_dataloader = DataLoader(
            val_dataset,
            batch_size=cfg.model.device_bsz,
            shuffle=False,
            num_workers=4,
            drop_last=True,
            worker_init_fn=worker_init_fn,
        )

        # Initializing the Dataloaders might take time depending on process...
        xm.rendezvous("Initialized Dataloaders...")

        # Leverage the *special* <XLA ParallelLoader> API that handles synchronizing TPU cores across batches!
        #   > NOTE: This is super important!
        xm.master_print("\tSetting up Parallel MpDeviceLoaders...")
        train_device_loader = parallel.MpDeviceLoader(train_dataloader, device)
        val_device_loader = parallel.MpDeviceLoader(val_dataloader, device)

        # Book-keeping & LR setting when `resuming` --> only do this on start_epoch!
        if epoch == start_epoch:
            if start_checkpoint is not None:
                global_step = start_step + ((len(train_dataset) // cfg.model.effective_bsz) * start_epoch)
                resume_time = int(re.search("-t=(.+?).pt", str(start_checkpoint)).group(1))
                lrs.append(update_lr(start_epoch, start_step / (len(train_dataset) // cfg.model.effective_bsz)))
            else:
                lrs.append(update_lr(start_epoch, 0))

        # Iterate...
        step_start_time = time.time()
        with tqdm(total=len(train_device_loader) // accumulate_grad_batches, disable=not is_rank_zero) as progress:
            for train_idx, batch in enumerate(train_device_loader):
                if cfg.model.arch == "v-mvp":
                    # Run a forward pass through the MAE... other return vals are reconstructions (pixel norm) & mask
                    loss, _, _ = model(batch)
                    reconstruction_losses.append(loss)

                elif cfg.model.arch in {"v-r3m", "v-rn3m"}:
                    imgs, lang, lang_mask = batch
                    loss, tcn_loss, reward_loss, l1_loss, l2_loss, tcn_acc, rew_acc = model(imgs, lang, lang_mask)

                    # Add to trackers
                    tcn_losses.append(tcn_loss)
                    reward_losses.append(reward_loss)
                    l1_losses.append(l1_loss)
                    l2_losses.append(l2_loss)
                    tcn_accuracies.append(tcn_acc)
                    reward_accuracies.append(rew_acc)

                elif cfg.model.arch == "v-cond":
                    img, lang, lang_mask = batch
                    loss, _, _ = model(img, lang, lang_mask)
                    reconstruction_losses.append(loss)

                elif cfg.model.arch == "v-dual":
                    imgs, lang, lang_mask = batch
                    loss, [zero_loss, k_loss] = model(imgs, lang, lang_mask)

                    # Add to trackers
                    reconstruction_losses.append(loss)
                    zero_reconstruction.append(zero_loss)
                    k_reconstruction.append(k_loss)

                elif cfg.model.arch == "v-gen":
                    imgs, lang_con, lang_con_mask, lang_gen, lang_gen_mask, lang_gen_weight = batch
                    loss, reconstruction_loss, lm_loss, [zero_loss, k_loss] = model(
                        imgs, lang_con, lang_con_mask, lang_gen, lang_gen_mask, lang_gen_weight
                    )

                    # Add to trackers
                    reconstruction_losses.append(reconstruction_loss)
                    lm_losses.append(lm_loss)
                    lm_ppl.append(torch.exp(lm_loss))
                    zero_reconstruction.append(zero_loss)
                    k_reconstruction.append(k_loss)

                else:
                    raise NotImplementedError(f"Forward Pass Logic for Model `{cfg.model.arch}` not implemented!")

                # Write Loss to Loggers (prior to accumulation normalization)
                train_losses.append(loss)

                # Normalize loss to account for accumulation
                loss = loss / accumulate_grad_batches
                loss.backward()

                # Gradient Accumulation =>> Note: skip any errant batches at the end...
                if (train_idx + 1) % accumulate_grad_batches == 0:
                    xm.optimizer_step(optimizer)  # Note call to xm.optimizer_step() -- has implicit mark_step!
                    optimizer.zero_grad()

                    # Add to `step_times`
                    step_times.append(time.time() - step_start_time)

                    # Logging --> Because there is no guarantee processes will be in sync, we need a `closure`
                    #   > Ref: https://pytorch.org/xla/release/1.11/index.html#torch_xla.core.xla_model.add_step_closure
                    if is_rank_zero and global_step % cfg.tracking.log_frequency == 0:
                        if cfg.model.arch == "v-mvp":
                            xm.add_step_closure(
                                log_vmvp_train_update,
                                args=(
                                    epoch,
                                    global_step,
                                    run_id,
                                    train_losses,
                                    lrs[-1],
                                    reconstruction_losses,
                                    step_times,
                                ),
                            )

                        elif cfg.model.arch == "v-r3m":
                            xm.add_step_closure(
                                log_vr3m_train_update,
                                args=(
                                    epoch,
                                    global_step,
                                    run_id,
                                    train_losses,
                                    lrs[-1],
                                    tcn_losses,
                                    reward_losses,
                                    l1_losses,
                                    l2_losses,
                                    tcn_accuracies,
                                    reward_accuracies,
                                    step_times,
                                ),
                            )

                        elif cfg.model.arch == "v-rn3m":
                            xm.add_step_closure(
                                log_vrn3m_train_update,
                                args=(
                                    epoch,
                                    global_step,
                                    run_id,
                                    train_losses,
                                    lrs[-1],
                                    tcn_losses,
                                    reward_losses,
                                    l1_losses,
                                    l2_losses,
                                    tcn_accuracies,
                                    reward_accuracies,
                                    step_times,
                                ),
                            )

                        elif cfg.model.arch == "v-cond":
                            xm.add_step_closure(
                                log_vcond_train_update,
                                args=(
                                    epoch,
                                    global_step,
                                    run_id,
                                    train_losses,
                                    lrs[-1],
                                    reconstruction_losses,
                                    step_times,
                                ),
                            )

                        elif cfg.model.arch == "v-dual":
                            xm.add_step_closure(
                                log_vdual_train_update,
                                args=(
                                    epoch,
                                    global_step,
                                    run_id,
                                    train_losses,
                                    lrs[-1],
                                    reconstruction_losses,
                                    zero_reconstruction,
                                    k_reconstruction,
                                    step_times,
                                ),
                            )

                        elif cfg.model.arch == "v-gen":
                            xm.add_step_closure(
                                log_vgen_train_update,
                                args=(
                                    epoch,
                                    global_step,
                                    run_id,
                                    train_losses,
                                    lrs[-1],
                                    reconstruction_losses,
                                    lm_losses,
                                    lm_ppl,
                                    zero_reconstruction,
                                    k_reconstruction,
                                    step_times,
                                ),
                            )

                        else:
                            raise NotImplementedError(f"Log Update for Model `{cfg.model.arch}` not implemented!")

                    # Increment Global Step _after_ logging!
                    global_step += 1

                    # Save checkpoint subject to *local_step = (train_idx + 1) // accumulate_grad_batches*
                    saver.save(
                        epoch=epoch,
                        is_local_step=True,
                        model=model,
                        optimizer=optimizer,
                        duration=int(time.time() - start_time) + resume_time,
                        local_step=start_step + ((train_idx + 1) // accumulate_grad_batches),
                    )

                    # Update LR every `accumulation_steps` iterations...
                    lrs.append(
                        update_lr(
                            epoch,
                            (start_step + ((train_idx + 1) // accumulate_grad_batches))
                            / (len(train_dataset) // cfg.model.effective_bsz),
                        )
                    )

                    # Reset `step_start_time`
                    step_start_time = time.time()

                    # Update `progress` each time we take a gradient step!
                    progress.update()

                # After each forward pass, mark a step, to compile XLA graph for a single forward pass!
                #   =>> Note :: this is important, with gradient accumulation, the graph can get massive otherwise!
                xm.mark_step()

            else:
                # Clear gradients and reset start step (regardless) at end of the loop
                optimizer.zero_grad()
                start_step = 0

        # Redundant, but Synchronous Validation Epoch...
        xm.master_print("Validating...")
        val_losses = []
        with torch.no_grad():
            for batch in tqdm(val_device_loader, disable=not is_rank_zero):
                if cfg.model.arch == "v-mvp":
                    loss, _, _ = model(batch)
                elif cfg.model.arch in {"v-r3m", "v-rn3m"}:
                    imgs, lang, lang_mask = batch
                    loss, _, _, _, _, _, _ = model(imgs, lang, lang_mask)
                elif cfg.model.arch == "v-cond":
                    img, lang, lang_mask = batch
                    loss, _, _ = model(img, lang, lang_mask)
                elif cfg.model.arch == "v-dual":
                    imgs, lang, lang_mask = batch
                    loss, _ = model(imgs, lang, lang_mask)
                elif cfg.model.arch == "v-gen":
                    imgs, lang_con, lang_con_mask, lang_gen, lang_gen_mask, lang_gen_weight = batch
                    loss, _, _, _ = model(imgs, lang_con, lang_con_mask, lang_gen, lang_gen_mask, lang_gen_weight)
                else:
                    raise NotImplementedError(f"Forward Pass Logic for Model `{cfg.model.arch} not implemented!")

                # Just append to val_losses...
                val_losses.append(loss)

            # Compute Val Loss & *mesh reduce* --> Why? :: the XLA people said so!
            val_loss = torch.stack(val_losses).mean().item()
            val_loss = xm.mesh_reduce("val_loss", val_loss, np.mean)  # All replicas should just return the same thing?

            # Logging --> add another `closure` for end-of-epoch cleanup --> compute `duration` as well...
            duration = int(time.time() - start_time) + resume_time
            if is_rank_zero:
                xm.add_step_closure(
                    log_epoch_end_update,
                    args=(
                        cfg.model.arch,
                        epoch,
                        global_step,
                        run_id,
                        duration,
                        train_losses,
                        val_loss,
                        lrs[-1],
                        step_times,
                    ),
                )

        # Save Checkpoint (at end of Epoch)
        saver.save(
            epoch=epoch + 1,
            is_local_step=False,
            model=model,
            optimizer=optimizer,
            duration=duration,
            train_loss=train_losses[-1].item(),
            val_loss=val_loss,
        )

    # Dump TPU Debugging Metrics...
    if is_rank_zero:
        with open("tpu-debug-metrics.log", "w") as f:
            f.write(met.metrics_report())

    # Exiting w/ Multiprocessing is a Nightmare... try to join?
    xm.master_print("...and that's all, folks!")
    xm.rendezvous("Cheers!")

    # Sleep for like 3 minutes... get W&B to finish syncing logs
    wandb.finish()
    time.sleep(150)


def mp_fn(_: int, cfg: PretrainConfig) -> None:
    torch.set_default_tensor_type("torch.FloatTensor")

    # Let's Start Pretraining!
    xpretrain(cfg)


@hydra.main(config_path=None, config_name="config")
def main(cfg: PretrainConfig) -> None:
    import torch_xla.distributed.xla_multiprocessing as xmp

    # Call XMP Spawn w/ the Config as the sole argument...
    xmp.spawn(mp_fn, args=(cfg,), nprocs=cfg.accelerator.num_accelerators, start_method="spawn")


if __name__ == "__main__":
    main()



"""
transforms.py

Default video/image transforms for Voltron preprocessing and training. Provides utilities for defining different scale
and crop transformations on a dataset-specific basis.

There are two key desiderata we ensure with the transforms:
    - Aspect Ratio --> We *never* naively reshape images in a way that distorts the aspect ratio; we crop instead!
    - Minimum Size --> We *never* upsample images; processing strictly reduces dimensionality!
"""
from functools import partial
from typing import Any, Callable, List, Tuple

import torch
from PIL import Image, ImageOps
from torchvision.transforms import Compose, ConvertImageDtype, Lambda, Normalize, Resize


# Simple Identity Function --> needs to be top-level/pickleable for mp/distributed.spawn()
def identity(x: torch.Tensor) -> torch.Tensor:
    return x.float()


def scaled_center_crop(target_resolution: int, frames: List[Image.Image]) -> Image.Image:
    # Assert width >= height and height >= target_resolution
    orig_w, orig_h = frames[0].size
    assert orig_w >= orig_h >= target_resolution

    # Compute scale factor --> just a function of height and target_resolution
    scale_factor = target_resolution / orig_h
    for idx in range(len(frames)):
        frames[idx] = ImageOps.scale(frames[idx], factor=scale_factor)
        left = (frames[idx].size[0] - target_resolution) // 2
        frames[idx] = frames[idx].crop((left, 0, left + target_resolution, target_resolution))

    # Return "scaled and squared" images
    return frames


def get_preprocess_transform(
    dataset_name: str, preprocess_resolution: int
) -> Callable[[List[Image.Image]], List[Image.Image]]:
    """Returns a transform that extracts square crops of `preprocess_resolution` from videos (as [T x H x W x C])."""
    if dataset_name == "sth-sth-v2":
        return partial(scaled_center_crop, preprocess_resolution)
    else:
        raise ValueError(f"Preprocessing transform for dataset `{dataset_name}` is not defined!")


def get_online_transform(
    dataset_name: str, model_arch: str, online_resolution: int, normalization: Tuple[Any, Any]
) -> Compose:
    """Returns an "online" torchvision Transform to be applied during training (batching/inference)."""
    if dataset_name == "sth-sth-v2":
        # Note: R3M does *not* expect normalized 0-1 (then ImageNet normalized) images --> drop the identity.
        if model_arch in {"v-r3m", "v-rn3m"}:
            return Compose([Resize((online_resolution, online_resolution), antialias=True), Lambda(identity)])
        else:
            return Compose(
                [
                    Resize((online_resolution, online_resolution), antialias=True),
                    ConvertImageDtype(torch.float),
                    Normalize(mean=normalization[0], std=normalization[1]),
                ]
            )
    else:
        raise ValueError(f"Online Transforms for Dataset `{dataset_name}` not implemented!")



"""
transforms.py

Default image/video transformations for various datasets.
"""
from typing import Any, Tuple

import cv2
import numpy as np
import torch
from torchvision.transforms import Compose, ConvertImageDtype, Lambda, Normalize


# Definitions of Video Transformations (Reference: `something-something-v2-baseline`)
class ComposeMix:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, imgs):
        for transformation, scope in self.transforms:
            if scope == "img":
                for idx, img in enumerate(imgs):
                    imgs[idx] = transformation(img)
            elif scope == "vid":
                imgs = transformation(imgs)
            else:
                raise ValueError("Please specify a valid transformation...")
        return imgs


class RandomCropVideo:
    def __init__(self, size):
        self.size = size

    def __call__(self, imgs):
        th, tw = self.size
        h, w = imgs[0].shape[:2]
        x1, y1 = np.random.randint(0, w - tw), np.random.randint(0, h - th)
        for idx, img in enumerate(imgs):
            imgs[idx] = img[y1 : y1 + th, x1 : x1 + tw]
        return imgs


class Scale:
    def __init__(self, size):
        self.size = size

    def __call__(self, img):
        return cv2.resize(img, tuple(self.size))


def identity(x):
    """Transform needs to be pickleable for multiprocessing.spawn()."""
    return x.float()


def get_pre_transform(dataset: str, resolution: int, scale_factor: float = 1.1) -> ComposeMix:
    """Defines a `pre` transform to be applied *when serializing the images* (first pass)."""
    if dataset == "sth-sth-v2":
        if scale_factor > 1:
            transform = ComposeMix(
                [
                    [Scale((int(resolution * scale_factor), int(resolution * scale_factor))), "img"],
                    [RandomCropVideo((resolution, resolution)), "vid"],
                ]
            )
        else:
            transform = ComposeMix(
                [
                    [Scale((int(resolution * scale_factor), int(resolution * scale_factor))), "img"],
                ]
            )

        return transform
    else:
        raise NotImplementedError(f"(Pre) transforms for dataset `{dataset}` not yet implemented!")


def get_online_transform(dataset: str, model_arch: str, normalization: Tuple[Any, Any]) -> Compose:
    """Defines an `online` transform to be applied *when batching the images* (during training/validation)."""
    if dataset == "sth-sth-v2":
        # Note: R3M does *not* expect normalized 0-1 (then ImageNet normalized) images --> drop the identity.
        if model_arch in {"v-r3m", "v-rn3m"}:
            return Compose([Lambda(identity)])
        else:
            return Compose([ConvertImageDtype(torch.float), Normalize(mean=normalization[0], std=normalization[1])])
    else:
        raise NotImplementedError(f"(Online) transforms for dataset `{dataset} not yet implemented!")



"""
datasets.py

Core Pytorch Dataset implementations for the various "data flavors" used by the different representation learning models
(Voltron and data-locked reproductions). Crucially, ach dataset loads from the corresponding serialized batch files that
define the exact data (and order of iterating through the data) to see during each epoch.

Notably, these serialized files control exactly what data is seen by *all* methods *across epochs*; using these files is
critical to reproducibility & comparisons.

The file contains logic for a "standard" Dataset; all files (batch index files, image/video/language files) are stored
on local disk, assuming storage conducive to fast random reads. For a "streaming" dataset (loading data directly from
GCP Buckets/Amazon S3), see `v1/stream_datasets.py`.
"""
from pathlib import Path
from typing import Any, Optional, Tuple

import h5py
import numpy as np
import torch
from torch.utils.data import Dataset
from torchvision.io import read_image
from torchvision.transforms import Compose

from voltron.preprocessing.transforms import get_online_transform


class PretrainDataset(Dataset):
    def __init__(self) -> None:
        super().__init__()
        self.epoch, self.h5, self.vid, self.states = 0, None, None, None
        self.index_path, self.language_path, self.language = None, None, None

    def hydrate(self, path: Path) -> None:
        # Create Open HDF5 Handle
        self.h5 = h5py.File(path, "r")
        self.vid, self.states = self.h5["vid"].asstr(), self.h5["states"].asstr()

        # Load Language Index
        if self.language_path is not None:
            self.language = torch.load(self.index_path / self.language_path)

    def set_epoch(self, epoch: int) -> None:
        self.epoch = epoch

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, ...]:
        raise NotImplementedError("PretrainDataset is an abstract class; should never be initialized directly!")

    def __len__(self) -> int:
        raise NotImplementedError("PretrainDataset is an abstract class; should never be initialized directly!")


class StateDataset(PretrainDataset):
    def __init__(self, epoch: int, index_path: Path, img_transform: Compose, is_val: bool = False) -> None:
        super().__init__()
        self.index_path, self.is_val, self.val_loaded = index_path, is_val, False
        self.epoch, self.img_transform, self.hdf5_path, self.n_examples = epoch, img_transform, None, None

        # === Retrieve Epoch Batches --> only call before/between epochs (as we create new DataLoaders) ===
        self.set_epoch(epoch)

    def set_epoch(self, epoch: int) -> None:
        # Load Validation Batches
        if self.is_val and not self.val_loaded:
            self.hdf5_path = self.index_path / "state" / "validation-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

            # Set `val_loaded`
            self.val_loaded = True

        # Load Train Batches
        elif not self.is_val:
            self.hdf5_path = self.index_path / "state" / f"train-epoch={epoch}-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

    def __getitem__(self, idx: int) -> torch.Tensor:
        """Return processed image frame as a Tensor."""
        if self.h5 is None:
            self.hydrate(self.hdf5_path)

        return self.img_transform(read_image(str(self.index_path.parent / self.states[idx][0])))

    def __len__(self) -> int:
        return self.n_examples


class StateLanguageDataset(PretrainDataset):
    def __init__(
        self,
        epoch: int,
        index_path: Path,
        img_transform: Compose,
        lang_dropout: Optional[float] = None,
        is_val: bool = False,
    ) -> None:
        super().__init__()
        self.index_path, self.is_val, self.val_loaded = index_path, is_val, False
        self.epoch, self.img_transform, self.hdf5_path, self.n_examples = epoch, img_transform, None, None
        self.lang_dropout, self.dropout_idxs = 0.0 if (lang_dropout is None) else lang_dropout, set()

        # Set Language Path
        self.language_path = "val-language-index.pt" if self.is_val else "train-language-index.pt"

        # === Retrieve Epoch Batches --> only call before/between epochs (as we create new DataLoaders) ===
        self.set_epoch(epoch)

    def set_epoch(self, epoch: int) -> None:
        # Load Validation Batches
        if self.is_val and not self.val_loaded:
            self.hdf5_path = self.index_path / "state+language" / "validation-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

            # Set `val_loaded`
            self.val_loaded = True

        # Load Train Batches
        elif not self.is_val:
            self.hdf5_path = self.index_path / "state+language" / f"train-epoch={epoch}-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

        # Assemble Dropout Indices
        n_drop = int(self.lang_dropout * self.n_examples)
        self.dropout_idxs = set(np.random.choice(self.n_examples, n_drop, replace=False))

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Return processed image frame and language, decomposed as the input_ids and attention_mask."""
        if self.h5 is None:
            self.hydrate(self.hdf5_path)

        # Get Vid ID --> parse out language, transform frame!
        vid = self.vid[idx]
        lang, lang_mask = self.language[vid]["input_ids"], self.language[vid]["attention_mask"]

        # Dropout Language (Naive Zeroing leads to NaN --> just want the "CLS" token)
        if idx in self.dropout_idxs:
            # Initial language token is *always* <CLS> = `101` --> last token always <SEP> = `102`
            lang[1:] *= 0
            lang_mask[1:] *= 0

        # Retrieve Frame & Return
        img = self.img_transform(read_image(str(self.index_path.parent / self.states[idx][0])))
        return img, lang, lang_mask

    def __len__(self) -> int:
        return self.n_examples


class StateOKDataset(PretrainDataset):
    def __init__(
        self,
        epoch: int,
        index_path: Path,
        img_transform: Compose,
        lang_dropout: Optional[float] = None,
        is_val: bool = False,
    ) -> None:
        super().__init__()
        self.index_path, self.is_val, self.val_loaded = index_path, is_val, False
        self.epoch, self.img_transform, self.hdf5_path, self.n_examples = epoch, img_transform, None, None
        self.lang_dropout, self.dropout_idxs = 0.0 if (lang_dropout is None) else lang_dropout, set()

        # Set Language Path
        self.language_path = "val-language-index.pt" if self.is_val else "train-language-index.pt"

        # === Retrieve Epoch Batches --> only call before/between epochs (as we create new DataLoaders) ===
        self.set_epoch(epoch)

    def set_epoch(self, epoch: int) -> None:
        # Load Validation Batches
        if self.is_val and not self.val_loaded:
            self.hdf5_path = self.index_path / "state+ok" / "validation-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

            # Set `val_loaded`
            self.val_loaded = True

        # Load Train Batches
        elif not self.is_val:
            self.hdf5_path = self.index_path / "state+ok" / f"train-epoch={epoch}-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

        # Assemble Dropout Indices
        n_drop = int(self.lang_dropout * self.n_examples)
        self.dropout_idxs = set(np.random.choice(self.n_examples, n_drop, replace=False))

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Return processed dual frames and language, decomposed as the input_ids and attention_mask."""
        if self.h5 is None:
            self.hydrate(self.hdf5_path)

        # Get Vid ID --> parse out language, transform frames!
        vid = self.vid[idx]
        lang, lang_mask = self.language[vid]["input_ids"], self.language[vid]["attention_mask"]

        # Dropout Language (Naive Zeroing leads to NaN --> just want the "CLS" token)
        if idx in self.dropout_idxs:
            # Initial language token is *always* <CLS> = `101` --> last token always <SEP> = `102`
            lang[1:] *= 0
            lang_mask[1:] *= 0

        # Retrieve Frames & Return
        imgs = self.states[idx]
        imgs = torch.stack([self.img_transform(read_image(str(self.index_path.parent / fn))) for fn in imgs])
        return imgs, lang, lang_mask

    def __len__(self) -> int:
        return self.n_examples


class GenStateOKDataset(PretrainDataset):
    def __init__(
        self,
        epoch: int,
        index_path: Path,
        img_transform: Compose,
        gen_ratio: float,
        is_val: bool = False,
    ) -> None:
        super().__init__()
        self.index_path, self.is_val, self.val_loaded = index_path, is_val, False
        self.epoch, self.img_transform, self.hdf5_path, self.n_examples = epoch, img_transform, None, None
        self.gen_ratio, self.gen_idxs = gen_ratio, set()

        # Set Language Path
        self.language_path = "val-language-index.pt" if self.is_val else "train-language-index.pt"

        # === Retrieve Epoch Batches --> only call before/between epochs (as we create new DataLoaders) ===
        self.set_epoch(epoch)

    def set_epoch(self, epoch: int) -> None:
        # Load Validation Batches
        if self.is_val and not self.val_loaded:
            self.hdf5_path = self.index_path / "state+ok" / "validation-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

            # Set `val_loaded`
            self.val_loaded = True

        # Load Train Batches
        elif not self.is_val:
            self.hdf5_path = self.index_path / "state+ok" / f"train-epoch={epoch}-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

        # Assemble Generation Indices
        n_gen = int(self.gen_ratio * self.n_examples)
        self.gen_idxs = set(np.random.choice(self.n_examples, n_gen, replace=False))

    def __getitem__(
        self, idx: int
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, float]:
        """Return dual frames, conditioning language, language to generate, decomposed as input_ids/attention mask."""
        if self.h5 is None:
            self.hydrate(self.hdf5_path)

        # Get Vid ID --> parse out language to condition on / generate
        vid = self.vid[idx]
        lang_con, lang_con_mask = self.language[vid]["input_ids"], self.language[vid]["attention_mask"]
        lang_gen, lang_gen_mask, lang_gen_weight = lang_con.clone(), lang_con_mask.clone(), None

        # Generate / Condition Check --> (Naive Zeroing leads to NaN --> just want the "CLS" token)
        if idx in self.gen_idxs:
            # When Generating --> just condition on the <CLS> token and generate the rest!
            lang_con[1:] *= 0
            lang_con_mask[1:] *= 0
            lang_gen_weight = 1

        else:
            # When Conditioning -> just generate the <CLS> token (so things don't crash) but set weight to 0
            lang_gen[1:] *= 0
            lang_gen_mask[1:] *= 0
            lang_gen_weight = 0

        # Retrieve Frames and Return
        imgs = self.states[idx]
        imgs = torch.stack([self.img_transform(read_image(str(self.index_path.parent / fn))) for fn in imgs])

        return imgs, lang_con, lang_con_mask, lang_gen, lang_gen_mask, lang_gen_weight

    def __len__(self) -> int:
        return self.n_examples


class QuintetDataset(PretrainDataset):
    def __init__(self, epoch: int, index_path: Path, img_transform: Compose, is_val: bool = False) -> None:
        super().__init__()
        self.index_path, self.is_val, self.val_loaded = index_path, is_val, False
        self.epoch, self.img_transform, self.hdf5_path, self.n_examples = epoch, img_transform, None, None

        # Set Language Path
        self.language_path = "val-language-index.pt" if self.is_val else "train-language-index.pt"

        # === Retrieve Epoch Batches --> only call before/between epochs (as we create new DataLoaders) ===
        self.set_epoch(epoch)

    def set_epoch(self, epoch: int) -> None:
        # Load Validation Batches
        if self.is_val and not self.val_loaded:
            self.hdf5_path = self.index_path / "quintet+language" / "validation-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

            # Set `val_loaded`
            self.val_loaded = True

        # Load Train Batches
        elif not self.is_val:
            self.hdf5_path = self.index_path / "quintet+language" / f"train-epoch={epoch}-batches.hdf5"
            with h5py.File(self.hdf5_path, "r") as h5:
                self.n_examples = len(h5["states"])

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Return all five processed frames and language, decomposed as the input_ids/attention_mask."""
        if self.h5 is None:
            self.hydrate(self.hdf5_path)

        # Get Vid ID --> parse out language, transform frames!
        vid = self.vid[idx]
        lang, lang_mask = self.language[vid]["input_ids"], self.language[vid]["attention_mask"]

        # Retrieve Frames & Return
        imgs = self.states[idx]
        imgs = torch.stack([self.img_transform(read_image(str(self.index_path.parent / fn))) for fn in imgs])
        return imgs, lang, lang_mask

    def __len__(self) -> int:
        return self.n_examples


def get_datasets(
    epoch: int,
    dataset_name: str,
    model_arch: str,
    artifact_path: str,
    data_modality: str,
    resolution: int,
    normalization: Tuple[Any, Any],
    lang_dropout: Optional[float] = None,
    gen_ratio: Optional[float] = None,
) -> Tuple[PretrainDataset, PretrainDataset]:
    index = Path(artifact_path) / dataset_name / "index"
    img_transform = get_online_transform(dataset_name, model_arch, resolution, normalization)

    # Switch on `data_modality` --> differs based on `model_arch` (e.g., MVP --> img, V-Cond --> img, language)
    if data_modality == "state":
        train_ds = StateDataset(epoch, index, img_transform)
        val_ds = StateDataset(epoch, index, img_transform, is_val=True)

    elif data_modality == "state+language":
        train_ds = StateLanguageDataset(epoch, index, img_transform, lang_dropout)
        val_ds = StateLanguageDataset(epoch, index, img_transform, lang_dropout, is_val=True)

    elif data_modality == "state+ok":
        # V-Dual --> don't return language modeling elements (causal attention mask, suffix language, etc.)
        if gen_ratio is None:
            train_ds = StateOKDataset(epoch, index, img_transform, lang_dropout)
            val_ds = StateOKDataset(epoch, index, img_transform, lang_dropout, is_val=True)

        # V-Gen --> add language modeling elements!
        else:
            train_ds = GenStateOKDataset(epoch, index, img_transform, gen_ratio)
            val_ds = GenStateOKDataset(epoch, index, img_transform, gen_ratio, is_val=True)

    elif data_modality == "quintet+language":
        train_ds = QuintetDataset(epoch, index, img_transform)
        val_ds = QuintetDataset(epoch, index, img_transform, is_val=True)

    else:
        raise ValueError(f"Data Modality `{data_modality}` is not supported!")

    return train_ds, val_ds



"""
process.py

Utility functions for preprocessing large-scale video/vision-language datasets in multiple passes, using multiprocessing
for parallelization. Exposes a three-phase sequence for preprocessing --> batching data:
    - Phase I (`extract_frames`): Read in raw (video clip, language) pairs, extract and serialize *all frames* to disk.

This script tries to be smart where it can, using multiprocessing.Pool in Phase I to speed up extraction; however, for
larger datasets YMMV. You might consider extracting the relevant logic, and using tools like SLURM Job Arrays, AWS
Lambda Functions, or GCP Cloud Run to "burst preprocess" data.
"""
import json
import logging
import multiprocessing as mp
import os
import shutil
from functools import partial
from pathlib import Path
from typing import Tuple

import torch
from rich.progress import track
from transformers import AutoTokenizer

from voltron.preprocessing.core import do_dry_run, process_clip, serialize_epoch
from voltron.preprocessing.transforms import get_preprocess_transform

# Grab Logger
overwatch = logging.getLogger(__file__)


def extract_frames(
    name: str,
    path: str,
    artifact_path: str,
    preprocess_resolution: int,
    n_val_videos: int,
    dry_run: bool = False,
) -> Tuple[Path, Path, Path, Path]:
    """Phase I: Extract and serialize *all frames* from video clips; uses multiprocessing to parallelize."""
    overwatch.info(f"Phase 1 Preprocessing :: Extracting Frames for Dataset `{name}`")

    # Overview of Return Values:
    #   `t_registry` and `v_registry` =>> store mappings of "video id" -> {metadata}
    #   `t_dir` and `v_dir` =>> store "processed data" (extracted frames)
    t_dir, v_dir = Path(artifact_path) / name / "train", Path(artifact_path) / name / "val"
    t_registry, v_registry = t_dir / "registry.json", v_dir / "registry.json"

    # Short-Circuit
    if t_registry.exists() and v_registry.exists():
        return t_registry, v_registry, t_dir, v_dir

    # Setup / Book-Keeping
    os.makedirs(t_dir, exist_ok=True)
    os.makedirs(v_dir, exist_ok=True)

    # Retrieve "pre-serialization" frame transform --> we scale down video frames (*while preserving aspect ratios*)
    #   and center crop each frame to `(preprocess_resolution, preprocess_resolution)`; saves on disk space (by a lot!)
    preprocess_transform = get_preprocess_transform(name, preprocess_resolution=preprocess_resolution)

    # Switch on dataset (`name`)
    if name == "sth-sth-v2":
        with open(Path(path) / "labels/train.json", "r") as f:
            annotations = json.load(f)
            train_ids, train_lang = [x["id"] for x in annotations], [x["label"] for x in annotations]

        with open(Path(path) / "labels/validation.json", "r") as f:
            annotations = json.load(f)[:n_val_videos]
            val_ids, val_lang = [x["id"] for x in annotations], [x["label"] for x in annotations]

    else:
        raise ValueError(f"Language/Metadata Extraction Pipeline for Dataset `{name}` not implemented!")

    # Run Dry-Run (if specified) --> single-threaded for debugging
    if dry_run:
        do_dry_run(name, path, train_ids, val_ids, preprocess_transform)

    # Otherwise =>> Iterate through all videos, dump all frames subject to the following structure:
    #   |-> .../processed/something-something-v2/
    #       |-> <split>/
    #           |-> <video-id>/frames<0..k>.jpg
    #
    # We'll build a single metadata file with a mapping <video-id> : ("language", n_frames)
    #   > To speed up serialization, we'll use a multiprocessing.Pool and max out CPU workers
    with mp.Pool(mp.cpu_count()) as pool:
        for k, save, vids, langs in [("train", t_dir, train_ids, train_lang), ("val", v_dir, val_ids, val_lang)]:
            overwatch.info(f"\tWriting `{k}` videos to disk...")

            # Spawn!
            process_fn, registration = partial(process_clip, name, Path(path), save, preprocess_transform), {}
            for key, value in track(
                pool.imap_unordered(process_fn, zip(vids, langs)),
                total=len(vids),
                transient=True,
            ):
                if key is not None:
                    registration[key] = value

            # Write Registration to Disk
            with open(t_registry if k == "train" else v_registry, "w") as f:
                json.dump(registration, f)

    # Return Paths to Registry & Extract Directories...
    return t_registry, v_registry, t_dir, v_dir


def preprocess_language(
    name: str,
    train_registry: Path,
    val_registry: Path,
    artifact_path: str,
    max_lang_len: int,
    language_model: str,
    hf_cache: str,
) -> Path:
    """Phase II: Iterate through Language Captions/Narrations and Normalize/Tokenize (truncate/pad to max length)."""
    overwatch.info(f"Phase 2 Preprocessing :: Normalizing & Tokenizing Language for Dataset `{name}`")
    t_index, v_index = train_registry.parent / "index.pt", val_registry.parent / "index.pt"
    t_json, v_json = train_registry.parent / "index.json", val_registry.parent / "index.json"
    index_dir = Path(artifact_path) / name / "index"
    os.makedirs(index_dir, exist_ok=True)

    # Short-Circuit
    if (index_dir / "train-language-index.json").exists() and (index_dir / "val-language-index.json").exists():
        return index_dir

    # Grab Language --> retain metadata for building index structures!
    with open(train_registry, "r") as f:
        train_metadata = json.load(f)
        train = [(vid, train_metadata[vid]["language"], train_metadata[vid]) for vid in train_metadata]

    with open(val_registry, "r") as f:
        val_metadata = json.load(f)
        val = [(vid, val_metadata[vid]["language"], val_metadata[vid]) for vid in val_metadata]

    # Assemble *all* language
    language = [x[1] for x in train + val]

    # Build AutoTokenizer (from `language_model` identifier)
    tokenizer = AutoTokenizer.from_pretrained(language_model, cache_dir=hf_cache)

    # If `max_lang_len` not specified, dump some statistics to compute...
    if max_lang_len == -1:
        # Naively tokenizes and pads to the "maximum length" of _all_ language... long tail is a problem!
        encoded_language = tokenizer(language, return_tensors="pt", padding=True)
        lengths = encoded_language["attention_mask"].sum(dim=1)

        # Compute a histogram of lengths
        hist = lengths.float().histc(bins=lengths.max()).int()
        overwatch.info(f"Histogram: {hist.numpy().tolist()}")
        raise AssertionError("Compute max length and update dataset configuration!")

    # Otherwise, we've already set the maximum length, so let's use it!
    overwatch.info(f"\tTokenizing all language in dataset to maximum length `{max_lang_len}`")
    encoded_language = tokenizer(
        language, return_tensors="pt", max_length=max_lang_len, truncation=True, padding="max_length"
    )
    input_ids, attention_mask = encoded_language["input_ids"], encoded_language["attention_mask"]
    train_input_ids, train_attention_mask = input_ids[: len(train)], attention_mask[: len(train)]
    val_input_ids, val_attention_mask = input_ids[len(train) :], attention_mask[len(train) :]

    # Assertion, just to sanity check
    assert len(val_input_ids) == len(val_attention_mask) == len(val), "Something went wrong tokenizing language..."

    # Compute `index.pt` contents
    overwatch.info("\tAssembling `train` and `val` index structures...")
    train_pt = {
        train[i][0]: {**train[i][2], **{"input_ids": train_input_ids[i], "attention_mask": train_attention_mask[i]}}
        for i in range(len(train))
    }
    val_pt = {
        val[i][0]: {**val[i][2], **{"input_ids": val_input_ids[i], "attention_mask": val_attention_mask[i]}}
        for i in range(len(val))
    }

    # Additionally dump JSON versions of the same --> downstream interpretability, XLA
    overwatch.info("JSONifying both Train and Validation Language")
    train_json, val_json = {}, {}
    for vid in track(train_pt, description="Train Language :: ", transient=True):
        train_json[vid] = {
            "language": train_pt[vid]["language"],
            "n_frames": train_pt[vid]["n_frames"],
            "input_ids": train_pt[vid]["input_ids"].numpy().tolist(),
            "attention_mask": train_pt[vid]["attention_mask"].numpy().tolist(),
        }

    for vid in track(val_pt, description="Validation Language :: ", transient=True):
        val_json[vid] = {
            "language": val_pt[vid]["language"],
            "n_frames": val_pt[vid]["n_frames"],
            "input_ids": val_pt[vid]["input_ids"].numpy().tolist(),
            "attention_mask": val_pt[vid]["attention_mask"].numpy().tolist(),
        }

    # Dump Structures...
    overwatch.info(f"Saving Torch indices to `{t_index}` and `{v_index}` respectively...")
    torch.save(train_pt, t_index)
    torch.save(val_pt, v_index)

    overwatch.info(f"Saving JSON indices to `{t_json}` and `{v_json}` respectively...")
    with open(t_json, "w") as f:
        json.dump(train_json, f)

    with open(v_json, "w") as f:
        json.dump(val_json, f)

    # Pull relevant files out into their own `index` directory...
    shutil.copy(t_json, index_dir / "train-language-index.json")
    shutil.copy(v_json, index_dir / "val-language-index.json")

    return index_dir


def unify_batches(
    name: str,
    train_registry: Path,
    val_registry: Path,
    train_dir: Path,
    val_dir: Path,
    index_dir: Path,
    batch_formats: Tuple[Tuple[str, Tuple[str, ...]], ...],
    max_epochs: int = 400,
    initial_final_alpha: float = 0.2,
) -> None:
    """Phase III: Assemble "Data-Locked" Batches for *all models* for *all epochs* for consistency!"""
    overwatch.info(f"Phase 3 Preprocessing :: Assembling *Data-Locked* Batches for Dataset `{name}`")

    # Load Registries
    with open(train_registry, "r") as f:
        train_registrations = json.load(f)

    with open(val_registry, "r") as f:
        val_registrations = json.load(f)

    # Assert last element of `batch_formats` assumes all prior subsets...
    full_set_inputs = set(batch_formats[-1][1])
    for _, subset_inputs in batch_formats[:-1]:
        assert full_set_inputs.issuperset(set(subset_inputs)), "We have a problem with batch formats..."

    # Assemble Tracking Data
    b_keys, unique_states = {b[0] for b in batch_formats}, set()

    # Parse out all "state"-specific Elements...
    state_elements = [s for s in full_set_inputs if "state_" in s]
    do_initial, do_final = "state_initial" in state_elements, "state_final" in state_elements
    n_int = len(state_elements) - 2 if ("state_initial" in state_elements and "state_final" in state_elements) else 0

    # Serialize Epochs
    overwatch.info("\tSerializing Epochs to JSON --> Storing mapping of Epoch -> Image Paths")
    for b in b_keys:
        os.makedirs(index_dir / b, exist_ok=True)

    # We only write the Validation Epoch once --> held constant across *all* of training!
    overwatch.info("\tWriting Validation Epoch to Disk")
    val_epoch_idx, _, uniq_s = serialize_epoch(
        index_dir,
        val_registrations,
        val_dir,
        batch_formats,
        do_initial,
        do_final,
        initial_final_alpha,
        n_int,
        epoch=0,
        is_validation=True,
    )

    # Update Trackers...
    if val_epoch_idx != -1:
        unique_states |= uniq_s

    # Compute length of epochs --> CPU Count should be no higher...
    epochs, n_frames_per_epoch = list(range(max_epochs)), -1

    # Parallelize Train Epoch Serialization
    overwatch.info("\tPlacing the Train Registry into Shared Memory")
    manager = mp.Manager()
    mg_registry = manager.dict(train_registrations)

    # Multiprocess --> the memory demands here are a bit higher, so limit workers by factor of 4
    with mp.Pool(mp.cpu_count() // 4) as pool:
        overwatch.info("\tWriting Train Batches per Epoch to Disk")
        precompute_fn = partial(
            serialize_epoch,
            index_dir,
            mg_registry,
            train_dir,
            batch_formats,
            do_initial,
            do_final,
            initial_final_alpha,
            n_int,
        )
        for epoch_idx, n_frames, uniq_s in pool.imap_unordered(precompute_fn, epochs):
            if epoch_idx == -1:
                continue

            # Update Trackers
            unique_states |= uniq_s
            n_frames_per_epoch = n_frames

    # Dump Statistics (Note :: Only makes sense on "initial" computation --> uninterrupted!)
    overwatch.info(f"Train Uniqueness: {len(unique_states)} States & {len(mg_registry)} Utterances")
    overwatch.info(f"Final Statistics :: 1 Epoch has ~ {n_frames_per_epoch} Frames...")



"""
pretrain.py

Core pretraining script for Native PyTorch (Single/Multi-) GPU pretraining on the Something-Something-v2 dataset; this
is basically just a 1-1 reproduction of the XLA pretraining script (`examples/xla-reference/xpretrain.py`) with just
a bit of cleanup, the default PyTorch DDP semantics (`torchrun`), using PyTorch 2.0.

Other notable differences from `xpretrain.py`:
    - Loads data from the local filesystem instead of streaming from a GCP bucket (can be added back easily!)
    - No TPU/XLA specific dependencies --> just PyTorch 2.0!

Run with:
    - [Single Node Multi-GPU ($K)]: `torchrun --standalone --nnodes 1 --nproc-per-node $K examples/pretrain/pretrain.py`
"""
import logging
import os
import re
import time
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import hydra
import torch
import torch.distributed as dist
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING, OmegaConf
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.data import DataLoader, DistributedSampler
from tqdm import tqdm

from voltron.conf import AcceleratorConfig, DatasetConfig, ModelConfig, TrackingConfig
from voltron.datasets import get_datasets
from voltron.models import get_model_optimizer
from voltron.overwatch import OverwatchRich
from voltron.util import CheckpointSaver, Metrics, ResumeableDistributedSampler, do_resume, set_global_seed

# Set Defaults (Hydra w/ Structured Configs)
DEFAULTS = [
    "_self_",
    {"model": "v-gen-base"},
    {"dataset": "sth-sth-v2"},
    {"accelerator": "torchrun"},
    {"tracking": "voltron-tracking"},
    {"override hydra/job_logging": "overwatch_rich"},
]


@dataclass
class PretrainConfig:
    # fmt: off
    defaults: List[Any] = field(default_factory=lambda: DEFAULTS)
    hydra: Dict[str, Any] = field(default_factory=lambda: {
        "run": {"dir": "runs/train/${model.identifier}+dataset-${dataset.name}"}
    })

    # Command Line Arguments
    run_id: Optional[str] = None                                        # Run ID for Logging
    seed: int = 21                                                      # Random Seed (for reproducibility)

    # Resume / Debug Behavior
    resume: bool = True                                                 # Whether to resume an existing run...
    wandb_resume_id: Optional[str] = None                               # W&B Run ID for `resume` behavior...

    # Composable / Structured Arguments
    model: ModelConfig = MISSING                                        # Model architecture for pretraining
    dataset: DatasetConfig = MISSING                                    # List of datasets for pretraining
    accelerator: AcceleratorConfig = MISSING                            # Accelerator (should always keep `torchrun`)
    tracking: TrackingConfig = MISSING                                  # Run/experiment tracking configuration
    # fmt: on


# Hydra Setup :: Retrieve ConfigStore (Singleton) & Register Components
cs = ConfigStore.instance()
cs.store(group="hydra/job_logging", name="overwatch_rich", node=OverwatchRich)
cs.store(name="config", node=PretrainConfig)


@hydra.main(config_path=None, config_name="config")
def pretrain(cfg: PretrainConfig) -> None:
    # Initialize Distributed Process Group --> assumes NCCL + Environment Variable Initialization (via `torchrun`)
    dist.init_process_group(backend="nccl", init_method="env://")
    device_id = dist.get_rank() % torch.cuda.device_count()
    is_rank_zero, rank, world_size = dist.get_rank() == 0, dist.get_rank(), dist.get_world_size()

    # Create Unique Run Name -- `resume = True` we assume the same "run_id"
    if cfg.run_id is None:
        cfg.run_id = run_dir = f"{cfg.model.identifier}+{cfg.dataset.name}-ddp-x{cfg.seed}"
    else:
        run_dir = cfg.run_id

    # Setup Logging (Rank 0 Only!) and Directory Handling
    overwatch = logging.getLogger(__file__)
    overwatch.setLevel(logging.INFO if is_rank_zero else logging.ERROR)
    overwatch.info("Voltron Training :: Assembling the Legendary Defender...")
    if is_rank_zero:
        os.makedirs(run_dir, exist_ok=True)

    # Let's Get Started!
    overwatch.info(
        '\t=>> "If you get too worried about what could go wrong, you might miss a chance to do something great."'
    )

    # Set Randomness & Get Dataloader `worker_init_fn` to ensure proper randomness in augmentations (if any)
    worker_init_fn = set_global_seed(cfg.seed, get_worker_init_fn=True)

    # Initialize Model & Optimizer --> Wrap in DDP / Device Handling
    #   > Note :: For (Standard) DDP Training --> initializing Optimizer before DDP == initializing after!
    overwatch.info("Initializing Model, Optimizer, and Learning Rate Scheduler")
    model, optimizer, update_lr = get_model_optimizer(cfg.model, cfg.dataset)
    model = DDP(model.to(device_id), device_ids=[device_id], output_device=device_id)

    # Handle Resume / Checkpoint Loading
    resume_checkpoint, resume_epoch, resume_step = do_resume(cfg.resume, run_dir=run_dir)
    if resume_checkpoint is not None:
        # IMPORTANT --> Load weights by mapping specifically to `cuda:<device_id>`!
        resume_state = torch.load(resume_checkpoint, map_location=f"cuda:{device_id}")
        model.load_state_dict(resume_state["model_state_dict"])
        optimizer.load_state_dict(resume_state["optimizer_state_dict"])

        dist.barrier()

    # Create Checkpoint Saver and Save Initial Checkpoint
    saver = CheckpointSaver(cfg.tracking.checkpoint_strategy, run_dir, is_rank_zero=is_rank_zero)
    if resume_checkpoint is None and resume_epoch == 0:
        overwatch.info("  | Saving 0th Epoch Checkpoint (Model Initialization)")
        saver.save(
            epoch=0, is_local_step=False, model=model, optimizer=optimizer, duration=0, train_loss=None, val_loss=None
        )
        dist.barrier()

    # Get Datasets --> Barrier after I/O Intensive Operation
    overwatch.info(f"Retrieving Dataset `{cfg.dataset.name}` prepared for Model `{cfg.model.arch}`")
    train_dataset, val_dataset = get_datasets(
        0,
        cfg.dataset.name,
        cfg.model.arch,
        cfg.dataset.artifact_path,
        cfg.model.data_modality,
        cfg.dataset.resolution,
        cfg.dataset.normalization,
        cfg.model.get("lang_dropout", None),
        cfg.model.get("gen_ratio", None),
    )
    dist.barrier()

    # Create Metrics =>> Handles on-the-fly computation, logging to JSONL and Weights & Biases
    metrics = Metrics(
        active_loggers=cfg.tracking.active_loggers,
        run_id=cfg.run_id,
        hparams=OmegaConf.to_container(cfg),
        model_arch=cfg.model.arch,
        is_rank_zero=is_rank_zero,
        tracking_cfg=cfg.tracking,
        tags=cfg.tracking.tags,
        resume=cfg.resume,
        resume_id=cfg.wandb_resume_id,
    )
    dist.barrier()

    # Configure Gradient Accumulation --> function of `effective_bsz`, `native_bsz`, and `WORLD_SIZE`
    assert cfg.model.effective_bsz % cfg.model.native_bsz == 0, "Device `native_bsz` must evenly divide `effective_bsz`"
    accumulate_grad_batches = cfg.model.effective_bsz // cfg.model.native_bsz // world_size
    overwatch.info(f"Running `{cfg.model.identifier}` Model Pretraining with Parameters =>")
    overwatch.info(f"  | Effective Batch Size = `{cfg.model.effective_bsz}`")
    overwatch.info(f"  | Per-Device Batch Size = `{cfg.model.native_bsz}`")
    overwatch.info(f"  | Distributed World Size = `{world_size}`")
    overwatch.info(f"  | Accumulation Steps = `{accumulate_grad_batches}`")

    # Start Train Loop --> Iterate through Epochs (Evaluation at end of Epoch)
    overwatch.info("Starting Training Loop")
    for epoch in range(resume_epoch, cfg.dataset.max_epochs):
        overwatch.info(f"  | [Epoch {epoch:03d}] Building Distributed Sampler & DataLoaders")
        train_dataset.set_epoch(epoch)
        dist.barrier()

        # [Custom] ResumeableDistributedSampler operates over *examples* --> start_step (full batches) * effective_bsz
        train_sampler = ResumeableDistributedSampler(
            seen_examples=resume_step * cfg.model.effective_bsz,
            resume_epoch=resume_epoch,
            dataset=train_dataset,
            num_replicas=world_size,
            rank=rank,
            shuffle=True,
            seed=cfg.seed,
        )
        train_sampler.set_epoch(epoch)
        val_sampler = DistributedSampler(val_dataset, num_replicas=world_size, rank=rank, shuffle=False, drop_last=True)

        # Create Epoch DataLoaders
        train_dl = DataLoader(
            train_dataset,
            batch_size=cfg.model.native_bsz,
            sampler=train_sampler,
            shuffle=False,
            num_workers=cfg.accelerator.num_workers,
            drop_last=True,
            pin_memory=True,
            prefetch_factor=4,
            worker_init_fn=worker_init_fn,
        )
        val_dl = DataLoader(
            val_dataset, batch_size=cfg.model.native_bsz, sampler=val_sampler, shuffle=False, num_workers=4
        )

        # Book-Keeping =>> Set LR when `resume = True` (or starting from scratch)
        if epoch == resume_epoch or epoch == 0:
            metrics.resume_time = (
                int(re.search("-t=(.+?).pt", str(resume_checkpoint)).group(1)) if resume_checkpoint is not None else 0
            )
            metrics.commit(
                global_step=resume_step + ((len(train_dataset) // cfg.model.effective_bsz) * resume_epoch),
                lr=update_lr(resume_epoch, resume_step / (len(train_dataset) // cfg.model.effective_bsz)),
                update_step_time=True,
            )

        # === Train Epoch ===
        model.train()
        status = metrics.get_status(epoch)
        overwatch.info(f"  | [Epoch {epoch:03d}] Running Train Loop")
        with tqdm(
            total=len(train_dl) // accumulate_grad_batches, desc=status, leave=False, disable=not is_rank_zero
        ) as progress:
            for train_idx, batch in enumerate(train_dl):
                # Model-Specific Handling
                if cfg.model.arch == "v-mvp":
                    img = batch
                    loss, _, _ = model(img.to(device_id, non_blocking=True))
                    metrics.commit(reconstruction_loss=loss)

                elif cfg.model.arch in {"v-r3m", "v-rn3m"}:
                    imgs, lang, lang_mask = batch
                    loss, tcn_loss, reward_loss, l1_loss, l2_loss, tcn_acc, rew_acc = model(
                        imgs.to(device_id, non_blocking=True),
                        lang.to(device_id, non_blocking=True),
                        lang_mask.to(device_id, non_blocking=True),
                    )
                    metrics.commit(
                        tcn_loss=tcn_loss,
                        reward_loss=reward_loss,
                        l1_loss=l1_loss,
                        l2_loss=l2_loss,
                        tcn_accuracy=tcn_acc,
                        reward_accuracy=rew_acc,
                    )

                elif cfg.model.arch == "v-cond":
                    img, lang, lang_mask = batch
                    loss, _, _ = model(
                        img.to(device_id, non_blocking=True),
                        lang.to(device_id, non_blocking=True),
                        lang_mask.to(device_id, non_blocking=True),
                    )
                    metrics.commit(reconstruction_loss=loss)

                elif cfg.model.arch == "v-dual":
                    imgs, lang, lang_mask = batch
                    loss, [zero_loss, k_loss] = model(
                        imgs.to(device_id, non_blocking=True),
                        lang.to(device_id, non_blocking=True),
                        lang_mask.to(device_id, non_blocking=True),
                    )
                    metrics.commit(
                        reconstruction_loss=loss,
                        zero_reconstruction_loss=zero_loss,
                        k_reconstruction_loss=k_loss,
                    )

                elif cfg.model.arch == "v-gen":
                    imgs, lang_con, lang_con_mask, lang_gen, lang_gen_mask, lang_gen_weight = batch
                    loss, reconstruction_loss, lm_loss, [zero_loss, k_loss] = model(
                        imgs.to(device_id, non_blocking=True),
                        lang_con.to(device_id, non_blocking=True),
                        lang_con_mask.to(device_id, non_blocking=True),
                        lang_gen.to(device_id, non_blocking=True),
                        lang_gen_mask.to(device_id, non_blocking=True),
                        lang_gen_weight,
                    )
                    metrics.commit(
                        reconstruction_loss=reconstruction_loss,
                        zero_reconstruction_loss=zero_loss,
                        k_reconstruction_loss=k_loss,
                        lm_loss=lm_loss,
                        lm_ppl=torch.exp(lm_loss),
                    )

                else:
                    raise ValueError(f"Forward() for Model `{cfg.model.arch}` is not implemented!")

                # Commit Loss (Prior to Normalization)
                metrics.commit(loss=loss)

                # Normalize Loss to account for Gradient Accumulation --> Backward!
                normalized_loss = loss / accumulate_grad_batches
                normalized_loss.backward()

                # Step =>> Check if done w/ Gradient Accumulation
                if (train_idx + 1) % accumulate_grad_batches == 0:
                    metrics.commit(update_step_time=True)

                    # Push Metrics every `log_frequency` steps...
                    if metrics.global_step % cfg.tracking.log_frequency == 0:
                        status = metrics.push(epoch)

                    # Optimizer Step --> Increment Global Step, Learning Rate, and Checkpoint (if specified)
                    optimizer.step()
                    optimizer.zero_grad()
                    lr = update_lr(
                        epoch,
                        (resume_step + ((train_idx + 1) // accumulate_grad_batches))
                        / (len(train_dataset) // cfg.model.effective_bsz),
                    )
                    metrics.commit(global_step=metrics.global_step + 1, lr=lr)
                    saver.save(
                        epoch,
                        is_local_step=True,
                        model=model,
                        optimizer=optimizer,
                        duration=int(time.time() - metrics.start_time) + metrics.resume_time,
                        local_step=resume_step + ((train_idx + 1) // accumulate_grad_batches),
                    )

                    # Update Progress Bar
                    progress.update()
                    progress.set_description(status)

        # === After Train Epoch --> Clear Gradients and reset `resume_step` ===
        optimizer.zero_grad()
        resume_step = 0

        # === Validation ===
        overwatch.info(f"  | [Epoch {epoch:03d}] Running Validation Loop")
        model.eval()

        # Accumulate `validation_losses` in order to `all_reduce` later!
        val_losses = []
        with torch.no_grad():
            for batch in tqdm(val_dl, disable=not is_rank_zero, leave=False):
                # Model-Specific Handling
                if cfg.model.arch == "v-mvp":
                    img = batch
                    val_loss, _, _ = model(img.to(device_id, non_blocking=True))

                elif cfg.model.arch in {"v-r3m", "v-rn3m"}:
                    imgs, lang, lang_mask = batch
                    val_loss, _, _, _, _, _, _ = model(
                        imgs.to(device_id, non_blocking=True),
                        lang.to(device_id, non_blocking=True),
                        lang_mask.to(device_id, non_blocking=True),
                    )

                elif cfg.model.arch == "v-cond":
                    img, lang, lang_mask = batch
                    val_loss, _, _ = model(
                        img.to(device_id, non_blocking=True),
                        lang.to(device_id, non_blocking=True),
                        lang_mask.to(device_id, non_blocking=True),
                    )

                elif cfg.model.arch == "v-dual":
                    imgs, lang, lang_mask = batch
                    val_loss, _ = model(
                        imgs.to(device_id, non_blocking=True),
                        lang.to(device_id, non_blocking=True),
                        lang_mask.to(device_id, non_blocking=True),
                    )

                elif cfg.model.arch == "v-gen":
                    imgs, lang_con, lang_con_mask, lang_gen, lang_gen_mask, lang_gen_weight = batch
                    val_loss, _, _, _ = model(
                        imgs.to(device_id, non_blocking=True),
                        lang_con.to(device_id, non_blocking=True),
                        lang_con_mask.to(device_id, non_blocking=True),
                        lang_gen.to(device_id, non_blocking=True),
                        lang_gen_mask.to(device_id, non_blocking=True),
                        lang_gen_weight,
                    )

                else:
                    raise ValueError(f"Forward() for Model `{cfg.model.arch}` is not implemented!")

                # Add to Validation Losses
                val_losses.append(val_loss)

        # All Reduce --> Push Epoch Metrics --> Checkpoint!
        validation_loss = torch.stack(val_losses).mean()
        dist.all_reduce(validation_loss)
        avg_val_loss = validation_loss / world_size
        if is_rank_zero:
            epoch_status, train_loss, training_duration = metrics.push_epoch(epoch, avg_val_loss)
            saver.save(
                epoch=epoch + 1,
                is_local_step=False,
                model=model,
                optimizer=optimizer,
                duration=training_duration,
                train_loss=train_loss.item(),
                val_loss=avg_val_loss.item(),
            )

        # === End of Epoch ===
        dist.barrier()

    # Finalize
    metrics.finalize()

    # And... we're done!
    overwatch.info("...and that's all, folks!")
    dist.barrier()


if __name__ == "__main__":
    # General Defaults --> should use Tensor Cores (kinda) if you have them!
    torch.set_float32_matmul_precision("high")
    torch.multiprocessing.set_start_method("spawn", force=True)

    pretrain()



"""
utils.py

Preprocessing utilities, including dry-run and single-video (single-example) processing. This file effectively defines
the "atomic" logic (take one video --> extract all frames, etc.), while the `process.py` functions invoke each unit
in a multiprocessing pool.
"""
import glob
import json
import logging
import os
import time
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

import av
import h5py
import numpy as np
import pandas as pd
from hurry.filesize import alternative, size
from PIL import Image
from rich.progress import track
from tqdm import tqdm

# Grab Logger
overwatch = logging.getLogger(__file__)
logging.getLogger("libav").setLevel(logging.ERROR)


# === General Utilities ===


# Videos are saved as `train_dir/{vid}/{vid}_idx={i}.jpg || if `relpath` then *relative path* `{split}/{vid}/...
def get_path(save_dir: Path, v: str, i: int, relpath: bool = False) -> str:
    return str((save_dir if not relpath else Path(save_dir.name)) / v / f"{v}_idx={i}.jpg")


# === Dry-Run Functionality ===


def do_dry_run(
    name: str,
    path: str,
    train_ids: List[str],
    val_ids: List[str],
    preprocess_transform: Callable[[List[Image.Image]], List[Image.Image]],
    n_train_videos: int = 1000,
    n_val_videos: int = 100,
    n_samples: int = 1000,
) -> None:
    """Iterates through a small subset of the total dataset, logs n_frames & average image size for estimation."""
    overwatch.info(f"Performing Dry-Run with {n_train_videos} Train Videos and {n_val_videos} Validation Videos")
    dry_run_metrics = {
        "n_frames": [],
        "jpg_sizes": [],
        "n_samples": n_samples,
        "time_per_example": [],
        "blank": str(Path(path) / "blank.jpg"),
    }

    # Switch on dataset (`name`)
    if name == "sth-sth-v2":
        for k, n_iter, vids in [("train", n_train_videos, train_ids), ("val", n_val_videos, val_ids)]:
            for idx in track(range(n_iter), description=f"Reading {k.capitalize()} Videos =>> ", transient=True):
                container = av.open(str(Path(path) / "videos" / f"{vids[idx]}.webm"))
                assert int(container.streams.video[0].average_rate) == 12, "FPS for `sth-sth-v2` should be 12!"
                try:
                    imgs = [f.to_image() for f in container.decode(video=0)]
                except (RuntimeError, ZeroDivisionError) as e:
                    overwatch.error(f"{type(e).__name__}: WebM reader cannot open `{vids[idx]}.webm` - continuing...")
                    continue
                container.close()

                # Apply `preprocess_transform`
                imgs = preprocess_transform(imgs)

                # Dry-Run Handling --> write a dummy JPEG to collect size statistics, dump, and move on...
                dry_run_metrics["n_frames"].append(len(imgs))
                while dry_run_metrics["n_samples"] > 0 and len(imgs) > 0:
                    img = imgs.pop(0)
                    img.save(str(dry_run_metrics["blank"]))
                    dry_run_metrics["jpg_sizes"].append(os.path.getsize(dry_run_metrics["blank"]))
                    dry_run_metrics["n_samples"] -= 1

        # Compute nice totals for "dry-run" estimate...
        total_clips = len(train_ids) + len(val_ids)

    else:
        raise ValueError(f"Dry Run for Dataset `{name}` not implemented!")

    # Compute aggregate statistics and gently exit...
    avg_size, avg_frames = np.mean(dry_run_metrics["jpg_sizes"]), int(np.mean(dry_run_metrics["n_frames"]))
    overwatch.info("Dry-Run Statistics =>>")
    overwatch.info(f"\t> A video has on average `{avg_frames}` frames at {size(avg_size, system=alternative)}")
    overwatch.info(f"\t> So - 1 video ~ {size(avg_frames * avg_size, system=alternative)}")
    overwatch.info(
        f"\t> With the full dataset of {total_clips} Train + Val videos ~"
        f" {size(total_clips * avg_frames * avg_size, system=alternative)}"
    )
    overwatch.info("Dry-Run complete, do what you will... exiting ✌️")

    # Remove dummy file...
    os.remove(dry_run_metrics["blank"])
    exit(0)


# === Atomic "Processing" Steps ===


def process_clip(
    name: str,
    path: Path,
    save: Path,
    preprocess_transform: Callable[[List[Image.Image]], List[Image.Image]],
    item: Tuple[str, str],
) -> Tuple[Optional[str], Optional[Dict[str, Any]]]:
    """Processes a single video clip and extracts/serializes all frames (as jpeg), returning the registry contents."""
    if name == "sth-sth-v2":
        vid, lang = item
        container, registration = av.open(str(Path(path) / "videos" / f"{vid}.webm")), {"language": lang, "n_frames": 0}
        assert int(container.streams.video[0].average_rate) == 12, "FPS for `sth-sth-v2` should be 12!"
        try:
            imgs = [f.to_image() for f in container.decode(video=0)]
        except (RuntimeError, ZeroDivisionError) as e:
            overwatch.error(f"{type(e).__name__}: WebM reader cannot open `{vid}.webm` - continuing...")
            return None, None
        container.close()

        # Book-Keeping
        os.makedirs(save / vid, exist_ok=True)
        registration["n_frames"] = len(imgs)

        # Short Circuit --> Writes are Expensive!
        if len(glob.glob1(save / vid, "*.jpg")) == len(imgs):
            return vid, registration

        # Apply `preprocess_transform` --> write individual frames, register, and move on!
        imgs = preprocess_transform(imgs)
        for idx in range(len(imgs)):
            imgs[idx].save(get_path(save, vid, idx))

        # Return title & registration
        return vid, registration

    else:
        raise ValueError(f"Clip Processing for Dataset `{name}` is not implemented!")


# ruff: noqa: C901
def serialize_epoch(
    index_dir: Path,
    registry: Dict[str, Any],
    vid_dir: Path,
    batch_formats: Tuple[Tuple[str, Tuple[str, ...]], ...],
    do_initial: bool,
    do_final: bool,
    initial_final_alpha: float,
    n_int: int,
    epoch: int,
    is_validation: bool = False,
) -> Tuple[int, int, Optional[Set[str]]]:
    index_file = "validation-batches.json" if is_validation else f"train-epoch={epoch}-batches.json"
    index_hdf5 = "validation-batches.hdf5" if is_validation else f"train-epoch={epoch}-batches.hdf5"

    # Short-Circuit
    if all([(index_dir / key / index_file).exists() for key, _ in batch_formats]):
        return -1, -1, None

    # Random seed is inherited from parent process... we want new randomness w/ each process
    np.random.seed((os.getpid() * int(time.time())) % 123456789)

    # Create Tracking Variables
    unique_states, batches = set(), {b: [] for b, _ in batch_formats}

    # Iterate through Registry --> Note we're using `tqdm` instead of `track` here because of `position` feature!
    for vid in tqdm(registry.keys(), desc=f"Epoch {epoch}", total=len(registry), position=epoch):
        # The initial/final states are sampled from the first [0, \alpha) and final 1-\alpha, 1] percent of the video
        n_frames = registry[vid]["n_frames"]
        initial_idx, final_idx = 0, n_frames - 1
        if do_initial:
            initial_idx = np.random.randint(0, np.around(n_frames * initial_final_alpha))

        if do_final:
            final_idx = np.random.randint(np.around(n_frames * (1 - initial_final_alpha)), n_frames)

        # Assertion --> initial_idx < final_idx - len(state_elements)
        assert initial_idx < final_idx - n_int, "Initial & Final are too close... no way to sample!"

        # Assume remaining elements are just random "interior" states --> sort to get ordering!
        sampled_idxs = np.random.choice(np.arange(initial_idx + 1, final_idx), size=n_int, replace=False)
        sampled_idxs = sorted(list(sampled_idxs))

        # Compile full-set "batch"
        retrieved_states = [get_path(vid_dir, vid, x, relpath=True) for x in [initial_idx, *sampled_idxs] + [final_idx]]

        # Add batch to index for specific batch_format key...
        batches[batch_formats[-1][0]].append({"vid": vid, "states": retrieved_states, "n_frames": n_frames})
        unique_states.update(retrieved_states)

        # Add all other batch formats to indices...
        for key, elements in batch_formats[:-1]:
            n_states = len([x for x in elements if "state_" in x])
            assert (n_states <= 2) or (
                n_states == len(retrieved_states)
            ), f"Strange value of n_states={n_states} > 2 and not equal to total possible of {len(retrieved_states)}"

            # States are all independent -- each of the retrieved states is its own example...
            if n_states == 1:
                for idx in range(len(retrieved_states)):
                    batches[key].append({"vid": vid, "state": retrieved_states[idx], "n_frames": n_frames})

            # OK-Context is the only "valid" context for n_states == 2
            elif n_states == 2:
                assert elements == ["state_initial", "state_i", "language"], "n_states = 2 but not 0K context?"

                # Append 0th state to each of the remaining sampled contexts (usually 2 or 4)... each pair is an example
                for idx in range(1, len(retrieved_states)):
                    batches[key].append(
                        {"vid": vid, "states": [retrieved_states[0], retrieved_states[idx]], "n_frames": n_frames}
                    )

            # We're treating the entire sequence of retrieved states as a single example (for TCN/R3M/Temporal Models)
            else:
                batches[key].append({"vid": vid, "states": retrieved_states, "n_frames": n_frames})

    # Write JSON Index directly to disk...
    for key in batches:
        with open(index_dir / key / index_file, "w") as f:
            json.dump(batches[key], f)

    # Write HDF5 Index directly to disk...
    for key, elements in batch_formats[:-1]:
        n_states = len([x for x in elements if "state_" in x])

        # Create HDF5 File
        df = pd.DataFrame(batches[key])
        h5 = h5py.File(index_dir / key / index_hdf5, "w")
        for k in ["vid", "n_frames"]:
            h5.create_dataset(k, data=df[k].values)

        # Handle "state(s)" --> (image path strings) --> add leading dimension (`n_states`)
        if n_states == 1:
            dfs = df["state"].apply(pd.Series)
            h5.create_dataset("states", data=dfs.values)

        else:
            dfs = df["states"].apply(pd.Series)
            h5.create_dataset("states", data=dfs.values)

        # Close HDF5 File
        h5.close()

    return epoch, len(batches["state"]), unique_states



"""
datasets.py

Base Hydra Structured Config for defining various pretraining datasets and appropriate configurations. Uses a simple,
single inheritance structure.
"""
from dataclasses import dataclass
from typing import Any, Tuple

from hydra.core.config_store import ConfigStore
from hydra.utils import to_absolute_path
from omegaconf import MISSING


@dataclass
class DatasetConfig:
    name: str = MISSING
    path: str = MISSING
    artifact_path: str = MISSING

    # Streaming Parameters (assumes fully preprocessed dataset lives at `stream_prefix/...`)
    #   =>> Deprecated as of `v2`
    stream: bool = True
    stream_prefix: str = "data/processed"

    # Dataset-Specific Parameters
    resolution: int = 224
    normalization: Tuple[Any, Any] = MISSING

    # For preprocessing --> maximum size of saved frames (assumed square)
    preprocess_resolution: int = MISSING

    # Validation Parameters
    n_val_videos: int = MISSING

    # Language Modeling Parameters
    language_model: str = "distilbert-base-uncased"
    hf_cache: str = to_absolute_path("data/hf-cache")

    # Maximum Length for truncating language inputs... should be computed after the fact (set to -1 to compute!)
    max_lang_len: int = MISSING

    # Dataset sets the number of pretraining epochs (general rule :: warmup should be ~5% of full)
    warmup_epochs: int = MISSING
    max_epochs: int = MISSING

    # Plausible Formats --> These are instantiations each "batch" could take, with a small DSL
    #   > Note: Assumes final element of the list is the "most expressive" --> used to back-off
    batch_formats: Any = (
        ("state", ("state_i",)),
        ("state+language", ("state_i", "language")),
        ("state+ok", ("state_initial", "state_i", "language")),
        ("quintet+language", ("state_initial", "state_i", "state_j", "state_k", "state_final", "language")),
    )

    # Preprocessing :: Frame-Sampling Parameters
    initial_final_alpha: float = 0.2


@dataclass
class SthSthv2Config(DatasetConfig):
    # fmt: off
    name: str = "sth-sth-v2"
    path: str = to_absolute_path("data/raw/sth-sth-v2")
    artifact_path: str = to_absolute_path("data/processed/sth-sth-v2")

    # Dataset Specific arguments
    normalization: Tuple[Any, Any] = (                              # Mean & Standard Deviation (default :: ImageNet)
        (0.485, 0.456, 0.406),
        (0.229, 0.224, 0.225),
    )

    # Sth-Sth-v2 Videos have a fixed height of 240; we'll crop to square at this resolution!
    preprocess_resolution: int = 240

    # Validation Parameters
    n_val_videos: int = 1000                                        # Number of Validation Clips (fast evaluation!)

    # Epochs for Dataset
    warmup_epochs: int = 20
    max_epochs: int = 400

    # Language Modeling Parameters
    max_lang_len: int = 20
    # fmt: on


# Create a configuration group `dataset` and populate with the above...
#   =>> Note :: this is meant to be extendable --> add arbitrary datasets & mixtures!
cs = ConfigStore.instance()
cs.store(group="dataset", name="sth-sth-v2", node=SthSthv2Config)



"""
metrics.py

Utility classes defining Metrics containers with model-specific logging to various endpoints (JSONL local logs, W&B).
"""
import os
import re
import time
from abc import ABC, abstractmethod
from collections import deque
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import jsonlines
import numpy as np
import torch
import wandb

from voltron.conf import TrackingConfig

# === Define Loggers (`Logger` is an abstract base class) ===


class Logger(ABC):
    def __init__(self, run_id: str, hparams: Dict[str, Any], is_rank_zero: bool = False) -> None:
        self.run_id, self.hparams, self.is_rank_zero = run_id, hparams, is_rank_zero

    @abstractmethod
    def write_hyperparameters(self) -> None:
        raise NotImplementedError("Logger is an abstract class!")

    @abstractmethod
    def write(self, global_step: int, metrics: Dict[str, Union[int, float]]) -> None:
        raise NotImplementedError("Logger is an abstract class!")

    def finalize(self) -> None:
        time.sleep(1)


class JSONLinesLogger(Logger):
    def write_hyperparameters(self) -> None:
        if not self.is_rank_zero:
            return

        # Only log if `is_rank_zero`
        with jsonlines.open(f"{self.run_id}.jsonl", mode="w", sort_keys=True) as js_logger:
            js_logger.write(
                {
                    "run_id": self.run_id,
                    "start_time": datetime.now().strftime("%m-%d-%H:%M"),
                    "hparams": self.hparams,
                }
            )

    def write(self, global_step: int, metrics: Dict[str, Union[int, float]]) -> None:
        if not self.is_rank_zero:
            return

        # Only log if `is_rank_zero`
        with jsonlines.open(f"{self.run_id}.jsonl", mode="a", sort_keys=True) as js_logger:
            js_logger.write(metrics)


class WeightsBiasesLogger(Logger):
    def __init__(
        self,
        run_id: str,
        hparams: Dict[str, Any],
        tracking_cfg: TrackingConfig,
        tags: List[str],
        resume: bool = False,
        resume_id: Optional[str] = None,
        is_rank_zero: bool = False,
    ) -> None:
        super().__init__(run_id, hparams, is_rank_zero)
        self.tracking_cfg, self.tags, self.resume, self.resume_id = tracking_cfg, tags, resume, resume_id
        self.path = Path(os.getcwd() if self.tracking_cfg.directory is None else self.tracking_cfg.directory)

        # Handle (Automatic) Resume if `resume = True`
        if self.resume and self.resume_id is None:
            wandb_path = self.path / "wandb"
            if wandb_path.exists() and any((wandb_path / "latest-run").iterdir()):
                # Parse unique `run_id` from the `.wandb.` file...
                wandb_fns = [f.name for f in (wandb_path / "latest-run").iterdir() if f.name.endswith(".wandb")]
                assert len(wandb_fns) == 1, f"There should only be 1 `.wandb.` file... found {len(wandb_fns)}!"

                # Regex Match on `run-{id}.wandb`
                self.resume_id = re.search("run-(.+?).wandb", wandb_fns[0]).group(1)

            elif wandb_path.exists():
                raise ValueError("Starting Training from Scratch with Preexisting W&B Directory; Remove to Continue!")

        # Call W&B.init()
        self.initialize()

    def initialize(self) -> None:
        """Run W&B.init on the guarded / rank-zero process."""
        if not self.is_rank_zero:
            return

        # Only initialize / log if `is_rank_zero`
        wandb.init(
            project=self.tracking_cfg.project,
            entity=self.tracking_cfg.entity,
            config=self.hparams,
            name=self.run_id,
            dir=self.path,
            tags=self.tags,
            notes=self.tracking_cfg.notes,
            resume="allow" if self.resume else False,
            id=self.resume_id,
        )

    def write_hyperparameters(self) -> None:
        if not self.is_rank_zero:
            return

        # Only log if `is_rank_zero`
        wandb.config = self.hparams

    def write(self, global_step: int, metrics: Dict[str, Union[int, float]]) -> None:
        if not self.is_rank_zero:
            return

        # Only log if `is_rank_zero`
        wandb.log(metrics, step=global_step)

    def finalize(self) -> None:
        wandb.finish()
        time.sleep(150)


# === Core Metrics Container :: Responsible for Initializing Loggers and Compiling/Pushing Metrics ===


class Metrics:
    def __init__(
        self,
        active_loggers: List[str],
        run_id: str,
        hparams: Dict[str, Any],
        model_arch: str,
        is_rank_zero: bool,
        tracking_cfg: Optional[TrackingConfig] = None,
        tags: Optional[List[str]] = None,
        resume: bool = False,
        resume_id: Optional[str] = None,
        window: int = 128,
    ) -> None:
        """High-Level Container Logic for Metrics Logging; logic defined for each model architecture!"""
        self.model_arch, self.is_rank_zero, self.window = model_arch, is_rank_zero, window

        # Initialize Loggers
        self.loggers = []
        for log_type in active_loggers:
            if log_type == "jsonl":
                logger = JSONLinesLogger(run_id, hparams, is_rank_zero=is_rank_zero)
            elif log_type == "wandb":
                logger = WeightsBiasesLogger(
                    run_id, hparams, tracking_cfg, tags, resume, resume_id, is_rank_zero=is_rank_zero
                )
            else:
                raise ValueError(f"Logger `{log_type}` is not defined!")

            # Add Hyperparameters --> Add to `self.loggers`
            logger.write_hyperparameters()
            self.loggers.append(logger)

        # Create Universal Trackers
        self.global_step, self.start_time, self.resume_time, self.step_start_time = 0, time.time(), 0, time.time()
        self.tracker = {
            "loss": deque(maxlen=self.window),
            "lr": [],
            "step_time": deque(maxlen=self.window),
        }

        # Create Model-Specific Trackers
        if self.model_arch == "v-mvp":
            self.tracker.update({"reconstruction_loss": deque(maxlen=self.window)})

        elif self.model_arch in {"v-r3m", "v-rn3m"}:
            self.tracker.update(
                {
                    "tcn_loss": deque(maxlen=self.window),
                    "reward_loss": deque(maxlen=self.window),
                    "l1_loss": deque(maxlen=self.window),
                    "l2_loss": deque(maxlen=self.window),
                    "tcn_accuracy": deque(maxlen=self.window),
                    "reward_accuracy": deque(maxlen=self.window),
                }
            )

        elif self.model_arch == "v-cond":
            self.tracker.update({"reconstruction_loss": deque(maxlen=self.window)})

        elif self.model_arch == "v-dual":
            self.tracker.update(
                {
                    "reconstruction_loss": deque(maxlen=self.window),
                    "zero_reconstruction_loss": deque(maxlen=self.window),
                    "k_reconstruction_loss": deque(maxlen=self.window),
                }
            )

        elif self.model_arch == "v-gen":
            self.tracker.update(
                {
                    "reconstruction_loss": deque(maxlen=self.window),
                    "zero_reconstruction_loss": deque(maxlen=self.window),
                    "k_reconstruction_loss": deque(maxlen=self.window),
                    "lm_loss": deque(maxlen=self.window),
                    "lm_ppl": deque(maxlen=self.window),
                }
            )

        else:
            raise ValueError(f"Metrics for Model `{self.model_arch}` are not implemented!")

    def itemize(self) -> Dict[str, torch.Tensor]:
        """Utility method for converting `deque[torch.Tensor] --> mean over Tensors."""
        return {
            k: torch.stack(list(v)).mean().item()
            for k, v in self.tracker.items()
            if k not in {"loss", "lr", "step_time"}
        }

    def log(self, global_step: int, metrics: Dict[str, Union[int, float]]) -> None:
        for logger in self.loggers:
            logger.write(global_step, metrics)

    def finalize(self) -> None:
        for logger in self.loggers:
            logger.finalize()

    def get_status(self, epoch: int, loss: Optional[torch.Tensor] = None) -> str:
        lr = self.tracker["lr"][-1] if len(self.tracker["lr"]) > 0 else 0
        if loss is None:
            return f"=>> [Epoch {epoch:03d}] Global Step {self.global_step:06d} =>> LR :: {lr:.6f}"

        # Otherwise, embed `loss` in status!
        return f"=>> [Epoch {epoch:03d}] Global Step {self.global_step:06d} =>> LR :: {lr:.6f} -- Loss :: {loss:.4f}"

    def commit(
        self,
        *,
        global_step: Optional[int] = None,
        resume_time: Optional[int] = None,
        lr: Optional[float] = None,
        update_step_time: bool = False,
        **kwargs,
    ) -> None:
        """Update all metrics in `self.tracker` by iterating through special positional arguments & kwargs."""
        if not self.is_rank_zero:
            return

        # Special Positional Arguments
        if global_step is not None:
            self.global_step = global_step

        if resume_time is not None:
            self.resume_time = resume_time

        if lr is not None:
            self.tracker["lr"].append(lr)

        if update_step_time:
            self.tracker["step_time"].append(time.time() - self.step_start_time)
            self.step_start_time = time.time()

        # Generic Keyword Arguments
        for key, value in kwargs.items():
            self.tracker[key].append(value.detach())

    def push(self, epoch: int) -> str:
        """Push current metrics to loggers with model-specific handling."""
        if not self.is_rank_zero:
            return

        loss = torch.stack(list(self.tracker["loss"])).mean().item()
        step_time, lr = np.mean(list(self.tracker["step_time"])), self.tracker["lr"][-1]
        status = self.get_status(epoch, loss)

        # Model-Specific Handling
        itemized = self.itemize()
        if self.model_arch == "v-mvp":
            self.log(
                self.global_step,
                metrics={
                    "Pretrain/Step": self.global_step,
                    "Pretrain/Epoch": epoch,
                    "Pretrain/V-MVP Train Loss": loss,
                    "Pretrain/Reconstruction Loss": itemized["reconstruction_loss"],
                    "Pretrain/Learning Rate": lr,
                    "Pretrain/Step Time": step_time,
                },
            )

        elif self.model_arch in {"v-r3m", "v-rn3m"}:
            self.log(
                self.global_step,
                metrics={
                    "Pretrain/Step": self.global_step,
                    "Pretrain/Epoch": epoch,
                    f"Pretrain/V-{'R3M' if self.model_arch == 'v-r3m' else 'RN3M'} Train Loss": loss,
                    "Pretrain/TCN Loss": itemized["tcn_loss"],
                    "Pretrain/Reward Loss": itemized["reward_loss"],
                    "Pretrain/L1 Loss": itemized["l1_loss"],
                    "Pretrain/L2 Loss": itemized["l2_loss"],
                    "Pretrain/TCN Accuracy": itemized["tcn_accuracy"],
                    "Pretrain/Reward Accuracy": itemized["reward_accuracy"],
                    "Pretrain/Learning Rate": lr,
                    "Pretrain/Step Time": step_time,
                },
            )

        elif self.model_arch == "v-cond":
            self.log(
                self.global_step,
                metrics={
                    "Pretrain/Step": self.global_step,
                    "Pretrain/Epoch": epoch,
                    "Pretrain/V-Cond Train Loss": loss,
                    "Pretrain/Reconstruction Loss": itemized["reconstruction_loss"],
                    "Pretrain/Learning Rate": lr,
                    "Pretrain/Step Time": step_time,
                },
            )

        elif self.model_arch == "v-dual":
            self.log(
                self.global_step,
                metrics={
                    "Pretrain/Step": self.global_step,
                    "Pretrain/Epoch": epoch,
                    "Pretrain/V-Dual Train Loss": loss,
                    "Pretrain/Reconstruction Loss": itemized["reconstruction_loss"],
                    "Pretrain/Zero Reconstruction Loss": itemized["zero_reconstruction_loss"],
                    "Pretrain/K Reconstruction Loss": itemized["k_reconstruction_loss"],
                    "Pretrain/Learning Rate": lr,
                    "Pretrain/Step Time": step_time,
                },
            )

        elif self.model_arch == "v-gen":
            self.log(
                self.global_step,
                metrics={
                    "Pretrain/Step": self.global_step,
                    "Pretrain/Epoch": epoch,
                    "Pretrain/V-Gen Train Loss": loss,
                    "Pretrain/Reconstruction Loss": itemized["reconstruction_loss"],
                    "Pretrain/Zero Reconstruction Loss": itemized["zero_reconstruction_loss"],
                    "Pretrain/K Reconstruction Loss": itemized["k_reconstruction_loss"],
                    "Pretrain/CLM Loss": itemized["lm_loss"],
                    "Pretrain/CLM Perplexity": itemized["lm_ppl"],
                    "Pretrain/LM Loss": itemized["lm_loss"],
                    "Pretrain/LM Perplexity": itemized["lm_ppl"],
                    "Pretrain/Learning Rate": lr,
                    "Pretrain/Step Time": step_time,
                },
            )

        else:
            raise ValueError(f"Metrics.push() for Model `{self.model_arch}` is not implemented!")

        return status

    def push_epoch(self, epoch: int, val_loss: torch.Tensor) -> Tuple[str, torch.Tensor, int]:
        """End-of-Epoch => Push accumulated metrics to loggers with model-specific handling."""
        if not self.is_rank_zero:
            return

        # Compute End-of-Epoch Specialized Metrics
        loss, step_time = torch.stack(list(self.tracker["loss"])).mean(), np.mean(list(self.tracker["step_time"]))
        lr, duration = self.tracker["lr"][-1], int(time.time() - self.start_time) + self.resume_time
        epoch_status = (
            f"[Epoch {epoch:03d}] Global Step {self.global_step:06d} =>> LR :: {lr:.6f} -- Loss :: {loss:.4f} "
            f"-- Val Loss :: {val_loss:.4f} -- Total Time (sec) :: {duration}"
        )

        # Log for Model
        p_arch = {
            "v-mvp": "MVP",
            "v-r3m": "R3M (ViT)",
            "v-rn3m": "R3M (RN)",
            "v-cond": "V-Cond",
            "v-dual": "V-Dual",
            "v-gen": "V-Gen",
        }[self.model_arch]
        self.log(
            self.global_step,
            metrics={
                "Pretrain/Step": self.global_step,
                "Pretrain/Epoch": epoch,
                "Pretrain/Training Duration": duration,
                f"Pretrain/{p_arch} Train Epoch Loss": loss.item(),
                f"Pretrain/{p_arch} Train Loss": loss.item(),
                f"Pretrain/{p_arch} Validation Loss": val_loss.item(),
                "Pretrain/Learning Rate": lr,
                "Pretrain/Step Time": step_time,
            },
        )

        return epoch_status, loss, duration



"""
utils.py

Preprocessing utilities, including functions for dry-runs and processing a single video (helpers for multiprocessing
calls down the lines).
"""
import glob
import json
import logging
import os
import sys
import time
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import av
import cv2
import numpy as np
from hurry.filesize import alternative, size
from rich.progress import track
from tqdm import tqdm

from voltron.preprocessing.v1.transforms import ComposeMix

# Grab Logger
overwatch = logging.getLogger(__file__)
logging.getLogger("libav").setLevel(logging.ERROR)


# Videos are saved as `train_dir/{vid}/{vid}_idx={i}.jpg
def get_path(save_dir: Path, v: str, i: int) -> str:
    return str(save_dir / v / f"{v}_idx={i}.jpg")


def do_dry_run(
    name: str,
    path: str,
    n_train_videos: int,
    n_val_videos: int,
    train_ids: List[str],
    val_ids: List[str],
    pre_transform: ComposeMix,
    n_samples: int = 1000,
) -> None:
    """Iterates through a small subset of the total dataset, logs n_frames & average image size for estimation."""
    dry_run_metrics = {
        "n_frames": [],
        "jpg_sizes": [],
        "n_samples": n_samples,
        "time_per_example": [],
        "blank": str(Path(path) / "blank.jpg"),
    }
    if name == "sth-sth-v2":
        for k, n_iter, vids in [("train", n_train_videos, train_ids), ("val", n_val_videos, val_ids)]:
            for idx in track(range(n_iter), description=f"Reading {k.capitalize()} Videos =>> ", transient=True):
                vid = vids[idx]
                container = av.open(str(Path(path) / "videos" / f"{vid}.webm"))
                try:
                    imgs = [f.to_rgb().to_ndarray() for f in container.decode(video=0)]
                except (RuntimeError, ZeroDivisionError) as e:
                    overwatch.error(f"{type(e).__name__}: WebM reader cannot open `{vid}.webm` - continuing...")
                    continue

                # Close container
                container.close()

                # Apply `pre_transform`
                imgs = pre_transform(imgs)

                # Dry-Run Handling --> write a dummy JPEG to collect size statistics, dump, and move on...
                dry_run_metrics["n_frames"].append(len(imgs))
                while dry_run_metrics["n_samples"] > 0 and len(imgs) > 0:
                    img = imgs.pop(0)
                    cv2.imwrite(str(dry_run_metrics["blank"]), cv2.cvtColor(img, cv2.COLOR_RGB2BGR))
                    dry_run_metrics["jpg_sizes"].append(os.path.getsize(dry_run_metrics["blank"]))
                    dry_run_metrics["n_samples"] -= 1

        # Compute nice totals for "dry-run" estimation
        total_clips = len(train_ids) + len(val_ids)

    else:
        raise NotImplementedError(f"Dry Run for Dataset `{name}` not yet implemented!")

    # Compute Aggregate Statistics and gently exit...
    avg_size, avg_frames = np.mean(dry_run_metrics["jpg_sizes"]), int(np.mean(dry_run_metrics["n_frames"]))
    overwatch.info("Dry-Run Statistics =>>")
    overwatch.info(f"\t> A video has on average `{avg_frames}` frames at {size(avg_size, system=alternative)}")
    overwatch.info(f"\t> So - 1 video ~ {size(avg_frames * avg_size, system=alternative)}")
    overwatch.info(
        f"\t> With the full dataset of {total_clips} Train + Val videos ~"
        f" {size(total_clips * avg_frames * avg_size, system=alternative)}"
    )
    overwatch.info("Dry-Run complete, do what you will... exiting ✌️")

    # Remove dummy file...
    os.remove(dry_run_metrics["blank"])
    sys.exit(0)


def process_video(
    name: str, path: Path, save: Path, pre_transform: ComposeMix, item: Tuple[str, str]
) -> Tuple[Optional[str], Optional[Dict[str, Any]]]:
    """Processes a single video file, dumps to series of image files, and returns the registry contents."""
    if name == "sth-sth-v2":
        # For sth-sth-v2, `item` corresponds to a single video clip, so just a tuple!
        vid, lang = item
        container, registration = av.open(str(Path(path) / "videos" / f"{vid}.webm")), {"language": lang, "n_frames": 0}
        try:
            imgs = [f.to_rgb().to_ndarray() for f in container.decode(video=0)]
        except (RuntimeError, ZeroDivisionError) as e:
            overwatch.error(f"{type(e).__name__}: WebM reader cannot open `{vid}.webm` - skipping...")
            return None, None

        # Close container
        container.close()

        # Book-keeping
        os.makedirs(save / vid, exist_ok=True)
        registration["n_frames"] = len(imgs)

        # Early exit (writes are expensive)
        if len(glob.glob1(save / vid, "*.jpg")) == len(imgs):
            return vid, registration

        # Apply `pre_transform` --> write individual frames, register, and return
        imgs = pre_transform(imgs)
        for i in range(len(imgs)):
            cv2.imwrite(get_path(save, vid, i), cv2.cvtColor(imgs[i], cv2.COLOR_RGB2BGR))

        # Return title & registration
        return vid, registration

    else:
        raise NotImplementedError(f"Process Video for Dataset `{name}` not yet implemented!")


# ruff: noqa: C901
def precompute_epoch(
    index_dir: Path,
    registry: Dict[str, Any],
    vid_dir: Path,
    batch_formats: Tuple[Tuple[str, Tuple[str, ...]], ...],
    do_initial: bool,
    do_final: bool,
    initial_final_alpha: float,
    n_int: int,
    epoch: int,
    is_validation: bool = False,
) -> Tuple[int, int, Optional[Set[str]]]:
    index_file = "validation-batches.json" if is_validation else f"train-epoch={epoch}-batches.json"

    # Short-Circuit
    if all([(index_dir / key / index_file).exists() for key, _ in batch_formats]):
        return -1, -1, None

    # Random seed is inherited from parent process... we want new randomness w/ each process
    np.random.seed((os.getpid() * int(time.time())) % 123456789)

    # Create Tracking Variables
    unique_states, batches = set(), {b: [] for b, _ in batch_formats}

    # Iterate through Registry...
    for vid in tqdm(registry.keys(), desc=f"Epoch {epoch}", total=len(registry), position=epoch):
        # The initial/final states are sampled from the first [0, \alpha) and final 1-\alpha, 1] percent of the video
        n_frames = registry[vid]["n_frames"]
        initial_idx, final_idx = 0, n_frames - 1
        if do_initial:
            initial_idx = np.random.randint(0, np.around(n_frames * initial_final_alpha))

        if do_final:
            final_idx = np.random.randint(np.around(n_frames * (1 - initial_final_alpha)), n_frames)

        # Assertion --> initial_idx < final_idx - len(state_elements)
        assert initial_idx < final_idx - n_int, "Initial & Final are too close... no way to sample!"

        # Assume remaining elements are just random "interior" states --> sort to get ordering!
        sampled_idxs = np.random.choice(np.arange(initial_idx + 1, final_idx), size=n_int, replace=False)
        sampled_idxs = sorted(list(sampled_idxs))

        # Compile full-set "batch"
        retrieved_states = [get_path(vid_dir, vid, x) for x in [initial_idx, *sampled_idxs] + [final_idx]]

        # Add batch to index for specific batch_format key...
        batches[batch_formats[-1][0]].append({"vid": vid, "states": retrieved_states, "n_frames": n_frames})
        unique_states.update(retrieved_states)

        # Add all other batch formats to indices...
        for key, elements in batch_formats[:-1]:
            n_states = len([x for x in elements if "state_" in x])
            assert (n_states <= 2) or (
                n_states == len(retrieved_states)
            ), f"Strange value of n_states={n_states} > 2 and not equal to total possible of {len(retrieved_states)}"

            # States are all independent -- each of the retrieved states is its own example...
            if n_states == 1:
                for idx in range(len(retrieved_states)):
                    batches[key].append({"vid": vid, "state": retrieved_states[idx], "n_frames": n_frames})

            # OK-Context is the only "valid" context for n_states == 2
            elif n_states == 2:
                assert elements == ["state_initial", "state_i", "language"], "n_states = 2 but not 0K context?"

                # Append 0th state to each of the remaining sampled contexts (usually 2 or 4)... each pair is an example
                for idx in range(1, len(retrieved_states)):
                    batches[key].append(
                        {"vid": vid, "states": [retrieved_states[0], retrieved_states[idx]], "n_frames": n_frames}
                    )

            # We're treating the entire sequence of retrieved states as a single example (for TCN/R3M/Temporal Models)
            else:
                batches[key].append({"vid": vid, "states": retrieved_states, "n_frames": n_frames})

    # Write JSON Index directly to disk...
    for key in batches:
        with open(index_dir / key / index_file, "w") as f:
            json.dump(batches[key], f)

    return epoch, len(batches["state"]), unique_states
